# =========================================================================
#
# primitives used in the main loop
#

# =========================================================================

proc adaptor_parse {line} {
	# debugMsg "ToParse: ($line)"
	if {[regexp {^[ \t]*([cC]_[a-zA-Z0-9_]+)([ \t].*)?} $line match cmd rest]} {
		set the_rest [string trim $rest]
		return [list 1 [list $cmd $the_rest]]
	} else {
		return [list 0]
	}
}

# =========================================================================
proc utlog_tofile_concrete {args} {
}

#copied from ../Driver/logging.tcl
proc utlog_tostring_event_suspension { ev susp } {
	set repr $ev
	if {[string compare $susp ""] != 0} {
		if {$susp} {
			set repr "(Suspension)"
			set repr "(Quiescence)"
		}
	}
	return $repr
}
# =========================================================================

proc adaptor_do_channels {kind_sug channel_sug} {
	set fields {}
	lappend fields [list kind $kind_sug]
	lappend fields [list channel $channel_sug]
	set result "A_CHANNELS [fields_join $fields]"
	puts "A_CHANNELS_START"
	puts "A_CHANNELS_END"
}

proc adaptor_do_pcos {kind_sug channel_sug pco_sug} {
	puts "A_PCOS_START"
	puts "A_PCOS_END"
}

proc adaptor_do_iokind {kind_sug channel_sug} {
	set fields {}
	lappend fields [list kind $kind_sug]
	lappend fields [list channel $channel_sug]
	set result "A_IOKIND [fields_join $fields]"
	puts $result
}

proc adaptor_do_input {event channel suspension} {
	set tr_event [uttrace_parse_event $event]
	set g [lindex $tr_event 0]
	set exps [lindex $tr_event 1]
	set pco_concval_timestamp [iut_put_input $g $exps]
	set pco [lindex $pco_concval_timestamp 0]
	set concval [lindex $pco_concval_timestamp 1]
	set timestamp [lindex $pco_concval_timestamp 2]
	set fields {}
	lappend fields [list event $event]
	lappend fields [list channel $channel]
	lappend fields [list pco $pco]
	lappend fields [list suspension $suspension]
	lappend fields [list concrete $concval]
	lappend fields [list timestamp $timestamp]
	set result "A_INPUT [fields_join $fields]"
	puts $result
}

proc adaptor_do_output {channel} {
	set output [iut_get_output]
	debugMsg "Output $output"
	switch -- [lindex $output 0] \
	  output {
		set pco [lindex $output 1]
		set g_exps [lindex $output 2]
		set g [lindex $g_exps 0]
		set exps [lindex $g_exps 1]
		set event [uttrace_unparse_event $g_exps]
		set fields {}
		lappend fields [list event $event]
		lappend fields [list channel $channel]
		lappend fields [list pco $pco]
		lappend fields [list suspension 0]
		# lappend fields [list concrete $concval]
		# lappend fields [list timestamp $timestamp]
		set result "A_OUTPUT [fields_join $fields]"
		puts $result
	} delta {
		set fields {}
		lappend fields [list channel $channel]
		lappend fields [list suspension 1]
		set result "A_OUTPUT [fields_join $fields]"
		puts $result
	} error {
		set result "A_ERROR"
		puts $result
		fatal "error from iut\n"
	}
}

# =========================================================================

proc get_timestamp {} {
	set seconds [clock seconds]
	return "$seconds <[clock format $seconds]>"
}

# =========================================================================

proc iut_put_input { g exp } {
	set pco [event2pco $g $exp]
	if {[string length $pco] != 0} {
		set message [event2gcivalue $g $exp]
		TorxGciSend $pco $message
	} else {
		send_error "Warning: skipping event (unknown pco)\n\t$g $exp\n"
		set message ""
	}
	set timestamp [get_timestamp]
	return [list $pco $message $timestamp]
}

#
# iut_get_output
# no parameters
# special GCI version needed
#
proc iut_get_output {} {
	if {[main_using_gci]} {
		return [gci_iut_get_output]
	} else {
		return [nogci_iut_get_output]
	}
}
proc gci_iut_get_output {} {
	global iutinfo maininfo timeout sidinfo output_times timeout_times
	set timeval -1
	if {[GciQueueIsEmpty]} {
		TorxGciStartTimer 1 [expr int($iutinfo(timeout) * 1000)] ms
		while {[GciQueueIsEmpty]} {
			TorxGciSnapshot
		}
		set timeval [TorxGciReadTimer 1]
		TorxGciCancelTimer 1
	}
	set queue_el [GciQueueHead]
	send_error "gci_iut_get_output: $queue_el\n"
	if {[llength $queue_el] == 1} {
		set result $queue_el
	} else {
		set result [list [lindex $queue_el 0]  pcoName [gcivalue2event [lindex $queue_el 1] ] ]
	}
	if {[string match [lindex $result 0] output]} {
	    lappend output_times $timeval
	    lappend timeout_times O
	    send_error "OutputTimes: $output_times\n"
	} else {
	    lappend timeout_times $timeval
	    lappend output_times T
	    send_error "TimeoutTimes: $timeout_times\n"
	}
	send_error "gci_iut_get_output-result: $result\n"
	return $result
}

proc nogci_iut_get_output {} {
	global iutinfo maininfo timeout sidinfo
	set timeout $iutinfo(timeout)
	#	exp_internal 1
	# send_error "$iutinfo(expect_command)\n"
	eval $iutinfo(expect_command)
	set timeout $maininfo(timeout)
	#	exp_internal 0
	send_error "nogci_iut_get_output-result: $result\n"
	return $result
}

# =========================================================================

# Random number generator taken from:
#
# Brent Welch, "Practical Programming in Tcl/Tk", 2nd edition, pages 78,79
#   (adapted from Don Libes, "Exploring Expect", 1995 and from
#                 Press et al. "Numerical Recipes in C", 1998
#   )
#
proc RandomInit {seed} {
	global randomSeed
	set randomSeed $seed
}

proc Random {} {
	global randomSeed
	set randomSeed [expr ($randomSeed*9301 + 49297) % 233280]
	return [expr $randomSeed/double(233280)]
}

proc RandomRange {range} {
	return [expr int ([Random]*$range)]
}

# =========================================================================
proc default_channel { side } {
	switch -- $side {
		  input {
			set channel in
		} output {
			set channel out
		} default {
			set channel ""
		}
	}
}

# =========================================================================
