# ==============================
# variable manipulation

proc var_reset {} {
	global out
	catch {unset out}
}
proc varclear_any_ctxt {ctxt type experiment} {
	global out
	if {[info exists out(vars,$type,$ctxt)]} {
		foreach var $out(vars,$type,$ctxt) {
			if {[info exists out(vars,$type,$var,$ctxt)]} {
				debugMsg var "varclear_any_ctxt: clearing ($ctxt)($type)($experiment): $var"
				set value $out(vars,$type,$var,$ctxt)
				catch {unset out(hash,values,$type,$var,$ctxt,$value)}
				catch {unset out(vars,$type,$var,$ctxt)}
				catch {unset out(hash,vars,$type,$var,$ctxt)}
			}
		}
		catch {unset out(vars,$type,$ctxt)}
#		catch {unset out(hash,ctxt,$ctxt)}
	}
}
proc varclear_mk_ctxt {ctxt experiment} {
	varclear_any_ctxt $ctxt mk $experiment
}
proc varclear_cfg_ctxt {ctxt experiment} {
	varclear_any_ctxt $ctxt cfg $experiment
}

proc ruleclear_mk_ctxt {ctxt experiment} {
	global out
	set s {}
	if {[info exists out(rules,mk,$ctxt)]} {
		catch {unset out(rules,mk,$ctxt)}
	}
}

proc var_subst {value} {
	global gen
	regsub -all {\$campaign} $value $gen(campaign) value
	if {[info exists gen(experiment)]} {
		regsub -all {\$experiment} $value  $gen(experiment) value
	}
	return $value
}

# type should be either mk or cfg
proc var_any {ctxt kind type var value} {
	global out
	debugMsg var "var_any ($ctxt)($kind)($type)($var)($value)"
	set value [var_subst $value]
	switch -- $kind {
		normal {
			set out(vars,$type,$var,$ctxt) $value
		}
		array {
			if {![info exists out(hash,values,$type,$var,$ctxt,$value)]} {
				lappend out(vars,$type,$var,$ctxt) $value
				set out(hash,values,$type,$var,$ctxt,$value) 1
			}
		}
	}
	if {![info exists out(hash,ctxt,$ctxt)]} {
		lappend out(ctxt) $ctxt
		set out(hash,ctxt,$ctxt) 1
	}
	if {![info exists out(hash,vars,$type,$var,$ctxt)]} {
		lappend out(vars,$type,$ctxt) $var
		set out(hash,vars,$type,$var,$ctxt) 1
	}
}
proc arrayvar_cfg {ctxt var name value} {
	var_any $ctxt normal cfg "$var $name" $value
}
proc var_cfg {ctxt var value} {
	var_any $ctxt normal cfg $var $value
}
proc arrayvar_mk {ctxt var value} {
	var_any $ctxt array mk $var $value
}
proc var_mk {ctxt var value} {
	var_any $ctxt normal mk $var $value
}

proc vargen_mk {experiment} {
	global out
	set s {}
	foreach c $out(ctxt) {
		set s [concat $s [vargen_mk_ctxt $c $experiment]]
	}
	return $s
}

proc vargen_mk_ctxt {ctxt experiment} {
	global out
	set s {}
	if {[info exists out(vars,mk,$ctxt)]} {
		#puts "===== mk vars $ctxt $experiment ======================"
		foreach f $out(vars,mk,$ctxt) {
			if {[info exists out(vars,mk,$f,$ctxt)]} {
				if {[llength $out(vars,mk,$f,$ctxt)] <= 1} {
					set v [join $out(vars,mk,$f,$ctxt) " "]
					lappend s "$f=$v"
				} else {
					lappend s "$f= \\"
					set l $out(vars,mk,$f,$ctxt)
					while {[llength $l] > 1} {
						set v [lindex $l 0]
						lappend s  "  $v \\"
						set l [lrange $l 1 end]
					}
					if {[llength $l] > 0} {
						set v [lindex $l 0]
						lappend s  "  $v"
					}
				}
			}
		}
		#puts "===== end mk vars $ctxt $experiment =================="
	}
	return $s
}
proc vargen_cfg {experiment} {
	global out
	set s {}
	foreach c $out(ctxt) {
		set s [concat $s [vargen_cfg_ctxt $c $experiment]]
	}
	return $s
}
proc vargen_cfg_ctxt {ctxt experiment} {
	global out
	set s {}
	if {[info exists out(vars,cfg,$ctxt)]} {
		# puts "===== cfg vars $experiment ======================"
		foreach f [lsort $out(vars,cfg,$ctxt)] {
			set v [join $out(vars,cfg,$f,$ctxt) " "]
			lappend s "$f $v"
		}
		# puts "===== end cfg vars $experiment =================="
	}
	return $s
}

# ==============================
# we don't append rules to a list of rules, but instead,
# we append the target to a list of targets (if not yet present),
# and store the rule in assoc array indexed by target.
# This allows later rules to 'update' earlier ones, because then
# simply the assoc array entry is overwritten.
# We need this overwriting for the inheritence to work.
proc rule_mk {ctxt target depends commands} {
	global out
	debugMsg var "rule_mk ($ctxt)($target)($depends)($commands)"
	set out(rules,mk,$ctxt,$target) [list $depends $commands]
	if {![info exists out(hash,mk,ctxt,$ctxt,$target)]} {
		lappend out(targets,mk,$ctxt) $target
		set out(hash,mk,ctxt,$ctxt,$target) 1
	}
	if {![info exists out(hash,ctxt,$ctxt)]} {
		lappend out(ctxt) $ctxt
		set out(hash,ctxt,$ctxt) 1
	}
}

proc rulegen_mk {experiment} {
	global out
	set s {}
	foreach c $out(ctxt) {
		set s [concat $s [rulegen_mk_ctxt $c $experiment]]
	}
	return $s
}
proc rulegen_mk_ctxt {ctxt experiment} {
	global out
	set s {}
	if {[info exists out(targets,mk,$ctxt)]} {
		# puts "===== mk rules $ctxt $experiment ======================"
		foreach target $out(targets,mk,$ctxt) {
			set f $out(rules,mk,$ctxt,$target)
			set depends [lindex $f 0]
			set commands [lindex $f 1]
			lappend s "$target: $depends"
			foreach c $commands {
				lappend s "\t$c"
			}
		}
		# puts "===== end mk rules $ctxt $experiment =================="
	}
	return $s
}

# ==============================

proc include_mk {ctxt filename} {
	global out
	debugMsg var "include_mk ($ctxt)($filename)"
	if {![info exists out(hash,includes,mk,ctxt,$ctxt,$filename)]} {
		lappend out(includes,mk,$ctxt) $filename
		set out(hash,includes,mk,ctxt,$ctxt,$filename) 1
	}
	if {![info exists out(hash,ctxt,$ctxt)]} {
		lappend out(ctxt) $ctxt
		set out(hash,ctxt,$ctxt) 1
	}
}

proc includegen_mk_ctxt {ctxt experiment} {
	global out
	global maininfo
	debugMsg gen "includegen_mk_ctxt ($ctxt)($experiment)"
	set s {}
	if {[info exists out(includes,mk,$ctxt)]} {
		puts "===== mk includes $ctxt $experiment ======================"
		foreach name $out(includes,mk,$ctxt) {
			lappend s "#inc{ $name"
			# try to find the file.
			# First try it just like that;
			# if that fails, try it from the installation libdir
			if {[file readable $name]} {
				set thefile $name
			} elseif {[file readable [file join $maininfo(libdir) $name]]} {
				set thefile [file join $maininfo(libdir) $name]
			} else {
				warningMsg "$experiment: $ctxt: cannot find include file $name"
				continue
			}
			if {[catch {open $thefile  r} fileId]} {
				warningMsg "$experiment: $ctxt: cannot include $name: $fileId"
			} else {
				foreach line [split [read $fileId] \n] {
					lappend s $line
				}
				if {[lindex $s end] == {}} {
					set s [lreplace $s end end]
				}
				close $fileId
			}
			lappend s "#inc} $name"
		}
		puts "===== end mk includes $ctxt $experiment =================="
	}
	return $s
}

# ==============================

proc getfield {group name field} {
	global config
	global outvar
	set v {}
	set res $v
	if {[info exists config(field,list,$group,$name,$field)]} {
		set v $config(field,list,$group,$name,$field)
		foreach el $v {
			if {([string compare $group $field] == 0) &&
			    ([string compare $name $el] == 0)} {
				debugMsg getfield "getfield ($group)($name)($field) skipping value: $el"
				continue
			}
			if {[info exists outvar($group,$field,$el)]} {
				# keep a 'flat' list: avoid creation of sublists
				# (the variable-value may be a list)
				foreach vv $outvar($group,$field,$el) {
					lappend res $vv
				}
			} else {
				lappend res $el
			}
		}
	} elseif  {[info exists config(field,list,$group,$name,base)]} {
		set res [getfield $group $config(field,list,$group,$name,base) $field]
	}
	debugMsg getfield "getfield ($group)($name)($field): $res"
	return $res
}

proc groupexists {group name} {
	global config
	if {[info exists config(field,list,$group,$name)]} {
		return 1
	} else {
		return 0
	}
}

# ==============================
# recursive gen_* invocation

proc array_gen {group value} {
	debugMsg gen "do array_gen ($group)($value)"
	foreach v $value {
		gen $group $v
	}
	debugMsg gen "done array_gen ($group)($value)"
}

# we handle the 'outermost' special, because for that
# one we do try to invoke pre_gen_ and post_gen_ routines;
# we do not do that for the base classes.
proc gen {group value {super ""}} {
	if {[string compare $super ""] == 0} {
		set name $value
	} else {
		set name $super
	}
	set n [split $name =]
	if {([llength $n] == 2) && ([string compare [lindex $n 0] product] == 0)} {
		dogen_product [lindex $n 0] [lindex $n 1] $super
	} else {
		dogen_noproduct $group $value $super
	}
}
proc dogen_noproduct {group value {super ""}} {
	if {[string compare $super ""] == 0} {
		set name $value
	} else {
		set name $super
	}
	if {[string compare [info procs pre_gen_[set group]] "" ] != 0} {
		debugMsg gen "do pre_gen_[set group] $name"
		if {![catch {pre_gen_[set group] $name} foo]} {
			debugMsg gen "done pre_gen_[set group] $name"
		} else {
			warningMsg "failed pre_gen_[set group] $value $name: $foo"
		}
	}
	dogen $group $value $super
	if {[string compare [info procs post_gen_[set group]] "" ] != 0} {
		debugMsg gen "do post_gen_[set group] $name"
		if {![catch {post_gen_[set group] $name} foo]} {
			debugMsg gen "done post_gen_[set group] $name"
		} else {
			warningMsg "failed post_gen_[set group] $value $name: $foo"
		}
	}
}
# (recursively) handle blocks and its variable(s)
proc dogen {group value {super ""}} {
	global config
	debugMsg gen "do dogen ($group)($value)($super)"
	if {[string compare $super ""] == 0} {
		set name $value
	} else {
		set name $super
	}
	set stop 0
	set var [getfield $group $value var]
	if {[string compare $var ""] != 0} {
		foreach v $var {
			if {![groupexists var $v]} {
				warningMsg "$group=$value: var=$v undefined"
				set stop 1
			}
		}
	}
	if {$stop} {
		warningMsg "$group=$value: errors"
	} else {
		dogen_vars $group $value $super $var
	}
}
proc dogen_vars {group value super vars} {
	dogen_vars_rec var $group $value $super $vars "" ""
}
proc dogen_product {group value super} {
	if {[string compare $super ""] == 0} {
		set name $value
	} else {
		set name $super
	}
	if {[groupexists product $value]} {
		debugMsg genproduct "dogen_product ($name)"
		foreach f {type foreach prefix template} {
			debugMsg genproduct "dogen_product handling ($f)"
			set v [getfield product $value $f]
			if {[string compare $v ""] != 0} {
				set $f $v
			} else {
				set $f ""
			}
		}
		set stop 0
		if {![string compare $type $group]} {
			warningMsg "$group=$value: group $group mismatch $group=$value->type $type"
			set stop 1
		}
		if {![groupexists $type $template]} {
			warningMsg "$group=$value: $type=$template undefined"
			set stop 1
		}
		foreach f $foreach {
			if {![groupexists foreach $f]} {
				warningMsg "$group=$value: foreach=$f undefined"
				set stop 1
			}
		}
		if {$stop} {
			warningMsg "$group=$value: errors"
		} else {
			do_gen_product $name $type $foreach $prefix $template
		}
	} else {
		warningMsg "$group=$value: undefined"
	}
}
proc do_gen_product {productname type loops prefix template} {
	set res [dogen_vars_rec foreach product $productname $template $loops $type $prefix]
}

proc dogen_vars_rec {var group value super vars type prefix} {
	global outvar
	debugMsg gen "dogen_vars_rec ($var)($group)($value)($super)($vars)($type)($prefix)"
	set res ""
	if {[llength $vars] > 0} {
		set el [lindex $vars 0]
		set rest [lrange $vars 1 end]
		gen $var $el $super
		set varvalue [getfield $var $el value]
		set varname [getfield $var $el name]
		set vartypes [getfield $var $el type]
		if {([string compare $varvalue ""] != 0) &&
		    ([string compare $varname ""] != 0) &&
		    ([string compare $vartypes ""] != 0)} {
		    set tuples {}
			set new 1
		    foreach vartype $vartypes {
				set vartypelist [split $vartype .]
				set typegroup [lindex $vartypelist 0]
				set typefield [lindex $vartypelist 1]
				lappend tuples [list $typegroup $typefield]
				if {[info exists outvar($typegroup,$typefield,$varname)]} {
					if {[string compare $outvar($typegroup,$typefield,$varname) $varvalue] != 0} {
						warningMsg "$group=$value variable exists: $vartype $varname; using old value: $outvar($typegroup,$typefield,$varname)); ignoring new value: $varvalue"
					}
					set new 0
				}
			}
			if {[string compare $var foreach] == 0} {
				foreach v $varvalue {
					foreach t $tuples {
						set outvar([lindex $t 0],[lindex $t 1],$varname) $v
					}
					set res [concat $res [dogen_vars_rec $var $group $value $super $rest $type $prefix-$v]]
					foreach t $tuples {
						unset outvar([lindex $t 0],[lindex $t 1],$varname)
					}
				}
			} else {
				if {$new} {
					foreach t $tuples {
						set outvar([lindex $t 0],[lindex $t 1],$varname) $varvalue
					}
				}
					dogen_vars_rec $var $group $value $super $rest $type $prefix
				if {$new} {
					foreach t $tuples {
						unset outvar([lindex $t 0],[lindex $t 1],$varname)
					}
				}
			}
		}
	} else {
		if {[string compare $var foreach] == 0} {
			gen $type $super $prefix
			set res $prefix
		} else {
			dodogen $group $value $super
			if {[string compare $super ""] != 0} {
				set res $super
			} else {
				set res $value
			}
		}
	}
	debugMsg genproduct "done dogen_vars_rec ($var)($group)($value)($super)($vars)($type)($prefix)"
	return $res
}
# (recursively) handle blocks and their base(s)
proc dodogen {group value {super ""}} {
	global config
	debugMsg gen "do dodogen ($group)($value)($super)"
	if {[string compare $super ""] == 0} {
		set name $value
	} else {
		set name $super
	}
	if {[info exists config(field,list,$group,$value,base)]} {
		set base $config(field,list,$group,$value,base)
		if {[info exists config(field,list,$group,$base)]} {
			dogen $group $base $name
		} else {
			warningMsg "cannot (dodo)gen group $group=$value->base=$base: $group=$base undefined"
		}
	}
	if {[string compare [info procs gen_[set group]] "" ] != 0} {
		debugMsg gen "do gen_[set group] $value $super"
		if {![catch {gen_[set group] $value $super} foo]} {
			debugMsg gen "done gen_[set group] $value $super"
		} else {
			warningMsg "failed gen_[set group] $value $super: $foo"
		}
	} elseif {[info exists config(field,list,$group,$value)]} {
		# warningMsg "failed to dodogen group $group=$value: $foo"
		foreach f $config(field,list,$group,$value) {
			if {[string compare $f base] == 0} {
			} elseif {[string compare $f var] == 0} {
			} elseif {[string compare [info procs gen_[set group]_[set f]] "" ] != 0} {
				set v [getfield $group $value $f]
				if {[string compare $f $group] == 0} {
					# take special care to avoid loops!
					set v [getfield $group $value $f]
					set ind [lsearch -exact $v $name]
					while {$ind >= 0} {
						set v [lreplace $v $ind $ind]
						set ind [lsearch -exact $v $name]
					}
					if {[llength $v] == 0} {
						# skip empty list
						continue
					}
				}
				debugMsg gen "do gen_[set group]_[set f] $name $v"
				if {[catch {gen_[set group]_[set f] $name $v} foo]} {
					warningMsg "failed to gen field $group=$value->$f=$v: $foo"
				} else {
					debugMsg gen "done gen_[set group]_$f $name $v"
				}
			}
		}
	} elseif {[config_has_children $group]} {
		warningMsg "$group=[set value]: cannot generate: no definition"
	}
	debugMsg gen "done dodogen ($group)($value)($super)"
}

# ==============================

proc genfile_append_anyvars_ctxt {type ctxt value} {
	global outfile
	debugMsg genfile_contents "genfile_append_anyvars_ctxt ($type)($ctxt){"
	debugMsg genfile_contents $value
	debugMsg genfile_contents "}"
	if {![info exists outfile(hash,$type,$ctxt)]} {
		set outfile(value,$type,$ctxt) $value
		set outfile(hash,$type,$ctxt) 1
		lappend outfile(ctxt,$type) $ctxt
	}
}

proc genfile_append_mkvars_ctxt {ctxt value} {
	genfile_append_anyvars_ctxt mkvars $ctxt $value
}
proc genfile_append_mkrules_ctxt {ctxt value} {
	genfile_append_anyvars_ctxt mkrules $ctxt $value
}
proc genfile_append_mkincludes_ctxt {ctxt value} {
	genfile_append_anyvars_ctxt mkincludes $ctxt $value
}
proc genfile_append_config_ctxt {ctxt value} {
	genfile_append_anyvars_ctxt config $ctxt $value
}

proc genfile_contents {ctxt types} {
	global outfile
	set res {}
	foreach t $types {
		foreach c [lsort $outfile(ctxt,$t)] {
			debugMsg genfile_contents "genfile_contents ($ctxt)($types):($t)($c): $outfile(value,$t,$c)"
#			if {($outfile(value,$t,$c) != {}) &&
#			    ((([llength $ctxt] == 0) && ($c == {})) ||
#				 ([lsearch -exact $ctxt $c] >= 0))} {
#           }
			if {($outfile(value,$t,$c) != {})} {
				if {([llength $ctxt] == 0) ||
 					([lsearch $ctxt $c] >= 0)} {
					lappend res [join $outfile(value,$t,$c) \n]
				}
           }
		}
	}
	debugMsg genfile_contents "genfile_contents ($ctxt)($types): {"
	debugMsg genfile_contents $res
	debugMsg genfile_contents "}"
	return $res
}
proc genfile_create {filename text} {
	global gen
	set rootlist [file split $gen(campaign)]
	set namelist  [file split $filename]
	set prefix ""
	while {[llength $rootlist] > 0} {
		if {[string compare [lindex $rootlist 0] [lindex $namelist 0]] != 0} {
			warningMsg "will not generate file outside campaign root ($gen(campaign)): $filename"
			return
		}
		set prefix [file join $prefix [lindex $rootlist 0]]
		set rootlist [lrange $rootlist 1 end]
		set namelist [lrange $namelist 1 end]
	}
	while {[llength $namelist] > 1} {
		set prefix [file join $prefix [lindex $namelist 0]]
		if {![file exists $prefix]} {
			verboseMsg "creating: $prefix"
			if {[catch {file mkdir $prefix} foo]} {
				warningMsg "could not create $prefix: $foo"
			}
		}
		set namelist [lrange $namelist 1 end]
	}
	if {[catch {open $filename w} fileId]} {
		warningMsg "Cannot open $filename: $fileId"
	} else {
		puts $fileId $text
		close $fileId
		verboseMsg "Created: $filename"
	}
}
# ==============================

proc component_add {type val} {
	global comps
	if {![info exists comps(hash,$type,$val)]} {
		set comps(hash,$type,$val) 1
		lappend comps(list) $type=$val
	}
}

proc component_list {} {
	global comps
	if {[info exists comps(list)]} {
		return $comps(list)
	} else {
		return {}
	}
}

proc component_reset {} {
	global comps
	catch {unset comps}
}
proc component_push {val} {
	global comps
	lappend comps(stack) $val
}
proc component_top {} {
	global comps
	if {[info exists comps(stack)]} {
		return [lindex $comps(stack) end]
	} else {
		return {}
	}
}
proc component_pop {} {
	global comps
	if {[info exists comps(stack)]} {
		set comps(stack) [lreplace $comps(stack) end end]
	}
}
# we assume that we have always at least the experiment on the stack
proc component_sfx {value} {
	global comps
	if {[llength $comps(stack)] > 1} {
		return " $value"
	} else {
		return ""
	}
}

# ==============================
