# =========================================================================
# $Id: fields.tcl,v 1.1.2.7 2001/04/05 09:18:04 feenstra Exp $
# =========================================================================

# args should be one of:
# - empty: return all config (list of tuples)
# - single name: return config for name (tuple)
# - name + value: set config for name (tuple)
proc fields_config {args} {
	global fieldcfg
	if {[llength $args] == 2} {
		set fieldcfg([lindex $args 0]) [lindex $args 1]
		return [list [lindex $args 0] $fieldcfg([lindex $args 0])]
	} elseif {[llength $args] == 1} {
		return [list [lindex $args 0] $fieldcfg([lindex $args 0])]
	} elseif {[llength $args] == 0} {
	} else {
	}	
}

proc fields_split {s {sep "\t"}} {
	set result [split [string trim $s] $sep]
#	puts stderr "fields_split ($s) ($sep): ($result)"
	return $result
}

proc fields_join {name_value_list} {
	set fields {}
	foreach nv $name_value_list {
		set n_escaped_v [list [lindex $nv 0] [field_escape [lindex $nv 1]]]
		lappend fields [join $n_escaped_v =]
	}
	return [join $fields "\t"]
}

proc field_value {items name {array 0}} {
	if {$array} {
		set result [field_value_array $items $name]
	} else {
		set result [field_value_item $items $name]
	}
#	puts stderr "field_value ($items) ($name) ($array) : ($result)"
	return $result
}

proc field_first {items} {
	if {[regexp "^(\[^=\]+)=(.*)" [lindex $items 0] all name value]} {
			return [list $name [field_un_escape $value]]
	}
	return ""
}

# =========================================================================

proc field_value_item {items name} {
	foreach i $items {
		if {[regexp "^$name=(.*)" $i all value]} {
			return [field_un_escape $value]
		}
	}
	return ""
}

proc field_value_array {items name} {
	set res {}
	foreach i $items {
		if {[regexp "^$name=(.*)" $i all value]} {
			lappend res [field_un_escape $value]
		}
	}
	return $res
}

# =========================================================================
# val is a string which might contain 'strings' that need to be replaced by
# other 'strings'. ("abcabcabcabc" {bc cb ab ba ca ac}) => "baaccbbaaccb"
#
# Assumption: in a single pass (non-recursive) string sustitutions take place in 
# order of occurrence: from left to right (val) and from head to tail (substlist).
# 
# hint: sort the substlist {pairwise!} in decreasing length and alfabeticaly.
# 
# 
proc string_substitution {val substlist} {
	set result ""
	if {[string length $val] > 0} {
		set curmatchidx [expr [string length $val]+1]
		set curmatch {}
		set cursubst {}
#		puts "val: $val"
#		puts "substlist: $substlist"
		foreach {xmatch xsubst} $substlist {
			set idx [string first $xmatch $val]
			if {$idx < 0} { continue }
			if {$idx < $curmatchidx} {
			   set curmatchidx $idx
			   set curmatch $xmatch
			   set cursubst $xsubst
#		puts "curmatchidx: $curmatchidx"
#		puts "curmatch: $curmatch"
#		puts "cursubst: $cursubst"
			}
		}
		if {[string length $curmatch] == 0} { return $val }
		if {$curmatchidx >= 0} {
			append result [string range $val 0 [expr $curmatchidx-1]]
			append result $cursubst
			set curmatchlen [string length $curmatch]
			set nextval [string range $val [expr $curmatchidx+$curmatchlen] end]
			append result [string_substitution $nextval $substlist]
		}
	}
	return $result
}

proc field_escape {val} {
	global fieldcfg
	if {[info exists fieldcfg(raw)] && $fieldcfg(raw)} {
		return $val
	} else {
		return [string_substitution $val {"\\\\" "\\\\\\\\" "\n" "\\n" "\r" "\\r" "\t" "\\t"}]
	}
}

proc field_un_escape {val} {
	global fieldcfg
	if {[info exists fieldcfg(raw)] && $fieldcfg(raw)} {
		return $val
	} else {
		return [string_substitution $val {"\\\\\\\\" "\\\\" "\\n" "\n" "\\r" "\r" "\\t" "\t"}]
	}
}

# =========================================================================
