# =========================================================================
#
# Error handling
#

proc do_stop {} {
	global maininfo
	set maininfo(stopnow) 1
}
proc get_pids {} {
	global adaptorinfo specinfo
	set pids {}
	foreach pidVar [list adaptorinfo(pid) specinfo(pid)] {
		if {[info exists $pidVar] &&
		   ![string match [set $pidVar] ""]} {
			lappend pids [set $pidVar]
		}
	}
	return $pids
}

proc will_clean_up {} {
	global maininfo
	set maininfo(cleaningup) 1
}
proc cleaning_up {} {
	global maininfo
	if {[info exists maininfo(cleaningup)] &&
	    $maininfo(cleaningup)} {
		return 1
	} else {
		return 0
	}
}

proc do_cleanup {} {
	global adaptorinfo specinfo sidinfo
	will_clean_up
	uttrace_file_close
	global sids
	set sids {}
	if {[info exists sidinfo(spec)]} {
		io_send_command spec QUIT
		io_receive_answer spec {
			-singleline {
				QUIT {} {
					catch {close -i $sidinfo{spec}}
					unset sidinfo(spec)
				}
			}
		}
	}
	if {[info exists sidinfo(adaptor)]} {
		io_send_command adaptor QUIT
		io_receive_answer adaptor {
			-singleline {
				QUIT {} {
					unset sidinfo(adaptor)
					catch {close -i $sidinfo{adaptor}}
				}
			}
		}
	}
	if {[info exists sidinfo(adaptor,stderr)]} {
		set timeout 3 ;# has to be at least 1 bigger than in adapter
		expect -i $sidinfo(adaptor,stderr) eof {
				set l $expect_out(buffer)
				send_error "ADAPTOR Stderr: $l\n"
				send_error "ADAPTOR Stderr Eof\n"
		}
	}
	utlog_file_close
}
proc clean_exit { {r 0} } {
	do_cleanup
	exit $r
}
proc to_user { prefix s } {
	# send_user "to_user: $prefix $s\n"
	set buf $s
	while {[regexp -- "^(\[^\r\n\]*)\r*\n(.*)$" $buf match line rest ]} {
		catch {send_error "$prefix$line\n" }
		set buf $rest
	}
	if {![string match $buf ""]} {
		catch {send_error "$prefix$buf\n" }
	}
}
proc do_output { prefix s } {
	set l [string length $s]
	if {$l != 0} {
		if {[string compare [string index $s [expr $l - 1]] "\n"] == 0} {
			catch {send_error "$prefix$s"}
		} else {
			catch {send_error "$prefix$s\n"}
		}
	}
}
proc fatal { {s ""} } {
	do_output "Fatal: " $s
	utlog_fatal [string trimright $s]
	will_clean_up
	set timeout 1
	catch { expect } ;# HACK to get rid of errors 'spawn id ... not open'
	clean_exit 1
}
### WARNING: 'error' is a Tcl command, so we don't
### use that; instead we use 'nonfatal'!!
proc nonfatal { {s ""} } {
	do_output "Error: " $s
}
proc warning { {s ""} } {
	do_output "Warning: " $s
}
proc message { {s ""} } {
	do_output "" $s
}
# don't use 'debug' as name of a proc, because
# expect already has a command 'debug'.
proc debugLvl { lvl {s ""} } {
	global maininfo
	if {$maininfo(debug) > $lvl} {
		do_output "DEBUG: " $s
	}
}

# =========================================================================
