# =========================================================================
#
# Logging routines
#

proc cygpath {p} {
	if {[catch {exec cygpath $p} newp]} {
		return $p
	} else {
		return $newp
	}
}

proc utlog_file_open {} {
	global maininfo
	if {[info exists maininfo(logfile)] &&
	    [string index $maininfo(logfile) 0] != "|"} {
		set logfile [cygpath $maininfo(logfile)]
		set name $logfile
		if {[file exists $logfile]} {
			set i 0
			set name $logfile.~$i~
			while {[file exists $name]} {
				incr i
				set name $logfile.~$i~
			}
			# if {[catch [exec mv $logfile $name] foo]} {
			# 	fatal "Cannot backup old logfile: $foo\n"
			# }
		}
		set maininfo(logfile) $name
		utlog_do_file_open $maininfo(logfile) file
	}
}

proc acceptIt {socket remoteHost remotePort} {
	global maininfo
	
	fconfigure $socket -buffering line -blocking 0
	fileevent $socket readable [list SockReader $socket]
	send_error "Log connection accepted from client tcp!$remoteHost!$remotePort at $socket\n"
	utlog_do_file_open $maininfo(log,name,newsockconnection) monitor "$remoteHost!$remotePort" $socket
}

proc SockReader { id } {
	global maininfo
	fileevent $id readable {}
	
	if {[eof $id]} {
		# "SockReader end of file reached"
		if {![catch {fconfigure $id -peername} cfg]} {
			set remoteHost [lindex $cfg 0]
			set remotePort [lindex $cfg 2]
			send_error "Log connection closed by client tcp!$remoteHost!$remotePort\n"
		}
		close $id
		set maininfo(log,file,ids) [ldelete $maininfo(log,file,ids) $id]
		send_error "Closed log $maininfo(log,type,$id) $maininfo(log,name,$id)\n"
		
		return
	}
	if {[gets $id line] < 0} {
		# blocked anyway because no full line available
		# dbgmsg "SockReader blocked"
		fileevent $id readable [list SockReader $id]
	} else {
		if {[process_sock_line $line $id]} {
			fileevent $id readable [list SockReader $id]
		}
	}
}

proc process_sock_line { s id } {
	global maininfo
	send_error "Log monitor $maininfo(log,name,$id): $s\n"
	# Terrible hack to capture tmcs output, it will look like:
	# Log monitor "tmcs": 0.0.0.0 0.0.0.0 34281
	if {[regexp tmcs $maininfo(log,name,$id)]} {
		set fields $s
		set maininfo(log,mon,mcastid) tcp![lindex $s 1]![lindex $s 2]
	}
	return 1
}

proc utlog_mon_open {defAddr defPort} {
	global maininfo sidinfo
	if {[info exists maininfo(logmon)] && $maininfo(logmon) != {}} {
		set nr 0
		match_max -d 10000
		# maininfo(logmon) contains a list of monitor commandlines
		# that may include arguments
		# send_error "about to spawn\n"
		spawn -noecho -pty
		# send_error "after spawn\n"
		set sidinfo(fifo,totorx) $spawn_id
		set totorxfifo $spawn_out(slave,name)
		set sidinfo(sid,name,$spawn_id) "LogMonitor"
		send_error "Created log monitor return pty $totorxfifo with spawn id $sidinfo(fifo,totorx)\n"
		send_error "DEBUG: log monitor match_max=[match_max]\n"
		expect_background {
			-i $sidinfo(fifo,totorx) full_buffer {
				global sidinfo
				warning "$sidinfo(sid,name,$expect_out(spawn_id)) Buffer Overflow!!!\n"
				set txt $expect_out(buffer)
				send_error "$sidinfo(sid,name,$expect_out(spawn_id)) Full Buffer, Contents Skipped: $txt\n"
				exit
			}
			-i $sidinfo(fifo,totorx) -re "(\[^\r\n]*)\r*\n" {
				global sidinfo
				set txt $expect_out(1,string)
				send_error "$sidinfo(sid,name,$expect_out(spawn_id)): $txt\n"
			}
		}
		
		send_error "Trying to create log monitor listen socket tcp!$defAddr!$defPort\n"
		if {$defAddr == {}} {
			set c [socket -server acceptIt $defPort]
		} else {
			set c [socket -server acceptIt -myaddr $defAddr $defPort]
		}
		set cfg [fconfigure $c -sockname]
		set ipNr [lindex $cfg 0]
		set portNr [lindex $cfg 2]
		send_error "Listening for log monitor connections on tcp!$ipNr!$portNr\n"
		
		# Start tmcs (tcp multicast service)
		set mline tmcs
		set cmd "exec torx-logclient $ipNr $portNr -- $mline > $totorxfifo 2>&1 &"
		set cmd [subst $cmd]
		set maininfo(log,name,newsockconnection) "\"$mline\""
		send_error "Trying to open log monitor: $cmd\n"
		if [catch {exec /bin/sh -c $cmd} msg] {
			warning "Cannot exec log monitor $mline: $msg\n"
		} else {
			vwait maininfo(log,file,ids)
			vwait maininfo(log,mon,mcastid)
			send_error "Found tmcs mcastid: $maininfo(log,mon,mcastid)\n"
			# TODO: connect to mcast, so that, if no-one else connects to it,
			#       it will go away when torx goes away.
			# set id [lindex $maininfo(log,file,ids) end]
			# if {[string compare $maininfo(log,name,$id) ""] == 0} {
			# 	set maininfo(log,name,$id) "\"$mline\""
			# }
			send_error "Opened $cmd\n"
		}
		
		
		# send_error "utlog_mon_open: before foreach\n"
		foreach mline $maininfo(logmon) {
			# deal with optional '# xtorx Tools menu entry' in mline
			# (from LOGMON entry) :
			#        command with args # xtorx Tools menu entry
			set ix [lsearch -exact $mline #]
			# puts stderr "$ix : $v"
			if {$ix >= 0} {
				set name [join [lrange $mline [expr $ix + 1] end] " "]
				set command [join [lrange $mline 0 [expr $ix - 1]] " "]
			} else {
				set name [file tail $mline]
				set command $mline
			}
			incr nr
			set cmd "torx-logclient"
			if {[string compare $maininfo(log,mon,mcastid) ""] != 0} {
				append cmd " -m $maininfo(log,mon,mcastid)"
			}
			append cmd " $ipNr $portNr -- $command > $totorxfifo 2>&1 &"
			set cmd [subst $cmd]
			set maininfo(log,name,newsockconnection) "\"$name\""
			send_error "Trying to open log monitor \"$name\": $cmd\n"
			if [catch {exec /bin/sh -c $cmd} msg] {
				warning "Cannot exec log monitor \"$name\" $command: $msg\n"
				continue
			} else {
				vwait maininfo(log,file,ids)
				# set id [lindex $maininfo(log,file,ids) end]
				# if {[string compare $maininfo(log,name,$id) ""] == 0} {
				# 	set maininfo(log,name,$id) "\"$mline\""
				# }
				send_error "Opened \"$name\" $cmd\n"
			}
		}
	}
}

proc utlog_do_file_open {name type {msg ""} {fid ""}} {
	global maininfo
	switch -- $type {
		file -
		fifo {
			if {[catch {open $name w} id]} {
				warning "Cannot open (for writing) log $type $name: $id\n"
				return {}
			} else {
				send_error "Opened log $type $name$msg\n"
				lappend maininfo(log,file,ids) $id
				set maininfo(log,name,$id) $name
				set maininfo(log,type,$id) $type
				return $id
			}
		} sid {
			if {[catch {exp_open -i $name} id]} {
				warning "Cannot open (for writing) log $type $name: $id\n"
				return {}
			} else {
				send_error "Opened log $type $name$msg as $id\n"
				lappend maininfo(log,file,ids) $id
				set maininfo(log,name,$id) $name
				set maininfo(log,type,$id) $type
				return $id
			}
		} monitor {
			set id $fid
			send_error "Opened log $type $name at $msg as $id\n"
			lappend maininfo(log,file,ids) $id
			set maininfo(log,name,$id) $name
			set maininfo(log,type,$id) $type
			return $id
		} default {
				warning "Internal error: do not know how to open log type: $type\n"
				return {}
		}
	}
}

proc utlog_file_close {} {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		utlog_retreive
		utlog_puts "VERDICT $maininfo(nrofstepsdone) $maininfo(verdict) $maininfo(verdict,guidance)"
		utlog_flush
		utlog_puts "EOF $maininfo(nrofstepsdone)"
		utlog_flush
		utlog_do_file_close
	}
	send_error "Verdict: $maininfo(nrofstepsdone) $maininfo(verdict) $maininfo(verdict,guidance)\n"
}
proc utlog_do_file_close {} {
	global maininfo sidinfo
	if {[info exists maininfo(log,file,ids)]} {
		foreach id $maininfo(log,file,ids) {
			if [catch {close $id} msg] {
				warning "Problems closing log $maininfo(log,type,$id) $maininfo(log,name,$id): $msg\n"
			} elseif {[string compare $maininfo(log,type,$id) fifo] == 0} {
				file delete $maininfo(log,name,$id)
			}
			set maininfo(log,file,ids) [ldelete $maininfo(log,file,ids) $id]
			send_error "Closed log $maininfo(log,type,$id) $maininfo(log,name,$id)\n"
		}
	}
	catch {close -i $sidinfo(fifo,totorx) }
}

proc ldelete { list value } {
	set ix [lsearch -exact $list $value]
	if {$ix >= 0} {
		return [lreplace $list $ix $ix]
	} else {
		return $list
	}
}

proc utlog_puts {s} {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		foreach id $maininfo(log,file,ids) {
			# puts stderr "Logging to $id: $s"
			if [catch {puts $id $s} err] {
				warning "Problems writing log $maininfo(log,type,$id) $maininfo(log,name,$id): $err\n"
				if {[string compare $maininfo(log,type,$id) fifo] == 0} {
					file delete $maininfo(log,name,$id)
				}
				set maininfo(log,file,ids) [ldelete $maininfo(log,file,ids) $id]
				send_error "Closed log $maininfo(log,type,$id) $maininfo(log,name,$id)\n"
			}
		}
	}
}

proc utlog_flush {} {
	global maininfo
	catch {unset maininfo(log,store)}
	if {[info exists maininfo(log,file,ids)]} {
		foreach id $maininfo(log,file,ids) {
			if [catch {flush $id} err] {
				warning "Problems flushing log $maininfo(log,type,$id) $maininfo(log,name,$id): $err\n"
				if {[string compare $maininfo(log,type,$id) fifo] == 0} {
					file delete $maininfo(log,name,$id)
				}
				set maininfo(log,file,ids) [ldelete $maininfo(log,file,ids) $id]
				send_error "Closed log $maininfo(log,type,$id) $maininfo(log,name,$id)\n"
			}
		}
	}
}

# routines utlog_store and utlog_retreive are used to temporarily
# store messages that
#  - will become redundant when newer info becomes available,
#  - but, if no new info is available by the time the log file is
#    closed, it really need only be written to logfile
#  - we suppose that utlog_flush will be used to flush the logfile,
#    when the 'new info' is written, so, after a utlog_flush,
#    we can forget what we stored.
#  - in utlog_close we invoke utlog_retreive

proc utlog_store {s} {
	global maininfo
	lappend maininfo(log,store) $s
}

proc utlog_retreive {} {
	global maininfo
	if {[info exists maininfo(log,store)]} {
		foreach s $maininfo(log,store) {
			utlog_puts $s
		}
		utlog_flush
	}
}

proc utlog_fatal {s} {
	utlog_retreive
	utlog_puts "FATAL $s"
	utlog_flush
}

proc utlog_tofile_configuration {} {
	global maininfo configinfo
	set configs_to_separate [list \
		INPUT           \
		OUTPUT          \
		PCO             \
		ADDRESS         \
		SPECFLAGS       \
		IUTFLAGS        \
		ADAPTORFLAGS    \
		IUTCONTEXTFLAGS \
		SOURCE          \
		LOGMON          \
	]
	if {[info exists maininfo(log,file,ids)]} {
		utlog_puts "CONFIGSTART"
		foreach k [lsort [array names configinfo]] {
			if {[string compare $k "unexpected"] == 0} {
				foreach l [lsort $configinfo($k)] {
					utlog_puts "CONFUNEXPECT $l"
				}
			} elseif { ([lsearch $configs_to_separate $k] >= 0) } {
				foreach l $configinfo($k) {
					utlog_puts "CONFIG $k $l"
				}
			} else {
				utlog_puts "CONFIG $k $configinfo($k)"
			}
		}
		utlog_puts "CONFIGEND"
		utlog_flush
	}
}

proc utlog_tofile_subconfiguration { role name configlist } {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		foreach c $configlist {
			utlog_puts "CONFSUB $role $name $c"
		}
		utlog_flush
	}
}

proc utlog_tostore { side chan ev susp preds } {
	global maininfo
	set pco ""
	if {[info exists maininfo(log,file,ids)]} {
		set chan_pco_opt [utlog_tostring_chan_pco $chan $pco]
		if {[string compare $susp ""] == 0} {
			set susp 0
		}
		# NOTE that side and chan_pco_opt should be next to each other
		# WITHOUT separating whitepace!!!
		utlog_store "ATTEMPT $maininfo(nrofstepsdone) $side$chan_pco_opt ($ev) ($susp)"
	}
}

proc utlog_tofile { side chan ev susp preds pco tmstmp conc } {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		set chan_pco_opt [utlog_tostring_chan_pco $chan $pco]
		if {[string compare $susp ""] == 0} {
			set susp 0
		}
		# NOTE that side and chan_pco_opt should be next to each other
		# WITHOUT separating whitepace!!!
		utlog_puts "ABSTRACT $maininfo(nrofstepsdone) $side$chan_pco_opt ($ev) ($susp)"
		# set res [join $conc ""]
		set res $conc
		regsub -all -- "\n" $res "\\n" res
		regsub -all -- "\r" $res "\\r" res
		utlog_puts "CONCRETE $maininfo(nrofstepsdone) $side $pco $chan ($res) ($tmstmp)"
		utlog_flush
	}
}

proc utlog_tofile_expectation { side tuple } {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		set chan [lindex $tuple 0]
		set ev [lindex $tuple 1]
		set susp [lindex $tuple 2]
		set preds [lindex $tuple 3]
		set p [string trim $preds]
		set repr $ev
		if {[string compare $p ""] != 0} {
			append repr " \[$p\]"
		}
		set pco ""
		set chan_pco_opt [utlog_tostring_chan_pco $chan $pco]
		# NOTE that side and chan_pco_opt should be next to each other
		# WITHOUT separating whitepace!!!
		utlog_puts "EXPECTED $maininfo(nrofstepsdone) $side$chan_pco_opt ($repr) ($susp)"
		utlog_flush
	}
}

proc utlog_tofile_refusal { side chan ev susp preds pco tmstmp conc  reason expl} {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		# NOTE that we print reason and explanation first,
		# such that when we see REFUSED
		# either we have seen explanation and reason,
		# or they are not there.
		if {[string compare $reason ""] != 0} {
			utlog_puts "REASON $maininfo(nrofstepsdone) $reason"
		}
		if {[string compare $expl ""] != 0} {
			utlog_puts "EXPLANATION $maininfo(nrofstepsdone) $expl"
		}
		set repr $ev
		if {[string compare $preds ""] != 0} {
			append repr " \[$preds\]"
		}
		set pco ""
		set chan_pco_opt [utlog_tostring_chan_pco $chan $pco]
		# NOTE that side and chan_pco_opt should be next to each other
		# WITHOUT separating whitepace!!!
		utlog_puts "REFUSED $maininfo(nrofstepsdone) $side$chan_pco_opt ($repr) ($susp)"
		utlog_flush
	}
}

proc utlog_tofile_timestamp {} {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		utlog_puts "TIMESTAMP $maininfo(nrofstepsdone) [clock seconds] [lindex [mstimer_cur] 1] <[clock format [clock seconds]]> [lindex [mstimer_cur] 1]"
		utlog_flush
	}
}

proc utlog_tofile_eospec {} {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		utlog_puts "LOG $maininfo(nrofstepsdone) EOSPEC"
		utlog_flush
	}
}

proc utlog_tofile_clicks_per_ms { factor } {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		utlog_puts "CLICKSPERMS $factor"
		utlog_flush
	}
}

proc utlog_tofile_stats { stats } {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		foreach stat $stats {
			utlog_puts "STATS $maininfo(nrofstepsdone) $stat"
		}
		utlog_flush
	}
}

proc utlog_tofile_stateid { stateid pfx } {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		foreach e $stateid {
			set str [join $e " "]
			utlog_puts "STATEID $maininfo(nrofstepsdone) $pfx $str"
		}
		utlog_flush
	}
}

proc utlog_tofile_string { who s } {
	global maininfo
	if {[info exists maininfo(log,file,ids)]} {
		utlog_puts "LOG $maininfo(nrofstepsdone) $who $s"
		utlog_flush
	}
}

proc utlog_tofile_mode {mode} {
	global maininfo
	if {[info exists maininfo(logfileId)]} {
		utlog_puts "MODE $maininfo(nrofstepsdone) $mode"
		utlog_flush
	}
}

proc utlog_tostring_chan_pco { chan pco } {
	if {[string compare $chan ""] != 0} {
		set chan_pco_opt "($chan"
		if {[string compare $pco ""] != 0} {
			append chan_pco_opt ":$pco"
		}
		append chan_pco_opt ")"
	} else {
		set chan_pco_opt ""
	}
	return $chan_pco_opt
}
proc utlog_tostring_event_suspension { side ev susp preds reason } {
	set repr $ev
	set p [string trim $preds]
	if {[string compare $p ""] != 0} {
		append repr " \[$p\]"
	}
	if {([string compare $susp ""] != 0) && $susp} {
		if {[string compare $side output] == 0} {
			set repr "(Suspension)"
			set repr "(Quiescence)"
		} elseif {[string compare $side input] == 0} {
			if {[string compare $reason ""] != 0} {
				set repr "(Refusal ($reason): $repr)"
			} else {
				set repr "(Refusal: $repr)"
			}
		}
	}
	return $repr
}
proc utlog_tostring_event_tuple { side chan ev susp preds pco reason } {
	set chan_pco_opt [utlog_tostring_chan_pco $chan $pco]
	set repr [utlog_tostring_event_suspension $side $ev $susp $preds $reason]
	set event "[set chan_pco_opt]: $repr"
	return $event
}

proc utlog_touser { side chan ev susp preds pco reason } {
	global maininfo loginfo
	set repr [utlog_tostring_event_tuple $side $chan $ev $susp $preds $pco $reason]
	set event "$maininfo(nrofstepsdone) [set side][set repr]"
	lappend loginfo(trace) $event
	send_user "$event\n"
}

proc utlog_trace {} {
	global loginfo
	if {[info exists loginfo(trace)]} {
		return $loginfo(trace)
	} else {
		return {}
	}
}

proc utlog_touser_menu { kind items } {
	foreach i $items {
		set chan [lindex $i 0]
		set ev [lindex $i 1]
		set susp [lindex $i 2]
		set preds [lindex $i 3]
		set repr [utlog_tostring_event_suspension $kind $ev $susp $preds ""]
		send_user "[set kind]($chan): $repr\n"
	}
}

proc utlog_touser_output { s } {
	global maininfo loginfo
	set l [string length $s]
	if {$l > 0} {
		if {[string compare [string index $s [expr $l - 1]] "\n"] == 0} {
			send_user "$s"
		} else {
			send_user "$s\n"
		}
	}
}

proc utlog_touser_error { s } {
	global maininfo loginfo
	set l [string length $s]
	if {$l > 0} {
		if {[string compare [string index $s [expr $l - 1]] "\n"] == 0} {
			send_user "$s"
		} else {
			send_user "$s\n"
		}
	}
}

# =========================================================================
