#! @PERL@ --

#
# usage: hexcontext [ -[no]printdata ] [ -[no]printdatahex ] [ -- ] program arguments
# 
# DECRIPTION:
#
# starts program with arguments,
# keeping a pipe to it.
# Then: read hex-lines from stdin and convert to binary and write to
# program, and read binary from stdout of program and convert to hex-lines.

use strict;

require 5.002;
use Getopt::Long;

my ( $prog, $BUFSIZ,
     $rin, $rout, $win, $wout, $ein, $eout,
     %optcl, $result, $debug, $printdata, $printdatahex,
     $command, $buffer, $bufferhex, %buf, $nfound, $nread,
     $message_hex, $message,
     $cpid, $cmd, @cmd, @fids,
   );

$BUFSIZ = 1024;

%optcl = ();
$result = GetOptions(
			\%optcl,
			"debug:i",
			"printdata!",
			"printdatahex!"
		);
die "usage: $0 [ -[no]printdata ] [ -[no]printdatahex ] [ -- ] [ program  [ args ]]\n"
	unless $result;

$debug = 1 if ((defined $optcl{"debug"}) && ! $optcl{"debug"});
$debug = $optcl{"debug"} if ((defined $optcl{"debug"}) && $optcl{"debug"});
$printdata = $optcl{"printdata"};
$printdatahex= 1 unless ((defined $optcl{"printdatahex"}) && ! $optcl{"printdatahex"});

if ($debug) {
#	print $prog.": ($result)(".$optcl{"debug"}.")(".$optcl{"printdata"}.")(".$optcl{"printdatahex"}.")\n";
#	print $prog.": ($result)(".$debug.")(".$printdata.")(".$printdatahex.")\n";
}
# exit if $debug >= 3;
# $debug =  3;

    pipe(ROFD,WOFD);    	# Open pipes that will link child and parent
    pipe(REFD,WEFD);
    pipe(RIFD,WIFD);
	my $oldfh;
	$oldfh = select WOFD; $| = 1; select $oldfh;
	$oldfh = select WEFD; $| = 1; select $oldfh;
	$oldfh = select WIFD; $| = 1; select $oldfh;
	$oldfh = select STDOUT; $| = 1; select $oldfh;
	$oldfh = select STDERR; $| = 1; select $oldfh;
    if (($cpid = fork()) == 0) {		# in child
		close(WIFD);				# close the unused end of pipes
		close(ROFD);
		close(REFD);
		close(STDOUT);				# attach stdout to pipe
		open(STDOUT,">&WOFD");
		close(STDERR);				# attach stderr to pipe
		open(STDERR,">&WEFD");
		close(STDIN);				# attach stdin to pipe
		open(STDIN,"<&RIFD");
		$| = 1;					# is this really needed?
		my $oldfh;
		$oldfh = select STDOUT; $| = 1; select $oldfh;
		$oldfh = select STDERR; $| = 1; select $oldfh;
		$oldfh = select WOFD; $| = 1; select $oldfh;
		$oldfh = select WEFD; $| = 1; select $oldfh;
		@cmd=@ARGV;	# construct say command line.
		exec @cmd;				# say it
		die "exec: prog: $!\n";			# or die
    }
    # in parent
    close(RIFD);				# close the unused end of pipes
    close(WOFD);
    close(WEFD);


$| = 1;				# don't buffer output on stdout
$oldfh = select STDERR; $| = 1; select $oldfh;

@fids = (fileno(STDIN), fileno(ROFD), fileno(REFD));

$rin = $win = $ein = "";
$rin = fdbits(@fids);
$ein = $rin | $win;

$buf{'STDIN'} = "";

if ($debug) {
	open(DEBUG, ">/dev/pts/2");
	$oldfh = select DEBUG; $| = 1; select $oldfh;
}
open(THESTDERR, ">&STDERR");
$oldfh = select THESTDERR; $| = 1; select $oldfh;

# we have to deail with (or ignore) the SIGHUP signal
# that will be sent when the adaptor closes the
# connection.
$SIG{'HUP'} = \&handler;
#$SIG{'INT'}  = \&handler;
#$SIG{'QUIT'} = \&handler;
#$SIG{'PIPE'} = \&handler;
#$SIG{'TERM'} = \&handler;
#	$SIG{TERM} = sub {close(WIFD); };
#	$SIG{HUP} = sub {close(WIFD); };
#	$SIG{INT} = sub {close(WIFD); };




for (;;) {
	print DEBUG "ERROR fileDescs:  ", &printVec($rin), "\n"
	    if $debug > 3;
	if (($nfound = select($rout=$rin, $wout=$win, $eout=$ein, undef)) < 0) {
	    warn "$prog: select: $!";
	}


	print DEBUG "rout:  ", &printVec($rout), ", " if $debug > 3;
	
	if (vec($eout, fileno(STDIN), 1)) {
		    print DEBUG "STDIN eof\n" if $debug;
			close(WIFD);
			@fids = handle_eof(fileno(STDIN));
			exit(0);
	}
	if (vec($rout, fileno(ROFD), 1)) {
		    print DEBUG "ERROR REFD eof\n" if $debug;
		    @fids = handle_eof(fileno(REFD));
	}
	if (vec($rout, fileno(STDIN), 1)) {
		print DEBUG "STDIN ready\n" if $debug > 2;
		$buffer = '';
		if (($nread = sysread(STDIN, $buffer, $BUFSIZ)) > 0) {
		    $buf{'STDIN'} .= $buffer;
		    print DEBUG "STDIN read $nread\n" if $debug > 2;
		} elsif ($nread == 0) {
		    print DEBUG "STDIN eof\n" if $debug;
			close(WIFD);
			@fids = handle_eof(fileno(STDIN));
		}
		if ($buf{'STDIN'} =~ /^([^\n]*)\n/) {
			$command = $1;
			$buf{'STDIN'} = $';
			if ($command =~ /^SENDHEX([ \t]+([0-9a-fA-F]+))?$/i) {
				$message_hex = $2;
			
				$message = pack("H*", $message_hex);
				print WIFD $message;
				print DEBUG "PROG send ".length($message)."\n" if $debug > 2;
			} elsif ($command =~ /^PRINTDATA$/i) {
				$printdata = 1;
			} elsif ($command =~ /^NOPRINTDATA$/i) {
				$printdata = 0;
			} elsif ($command =~ /^PRINTDATAHEX$/i) {
				$printdatahex = 1;
			} elsif ($command =~ /^NOPRINTDATAHEX$/i) {
				$printdatahex = 0;
			} elsif ($command =~ /^DEBUG([ \t]+([0-9]+))?$/i) {
				if ($2 ne "") {
					$debug = $2;
				} else {
					$debug = 1;
					# we still have to open the DEBUG file-handle...
				}
			} elsif ($command =~ /^NODEBUG$/i) {
				$debug = 0;
			}
		}
	}
	if (vec($rout, fileno(ROFD), 1)) {
		print DEBUG "ERROR ROFD ready\n" if $debug > 2;
		$buffer = '';
		$message = '';
		if (($nread = sysread(ROFD, $buffer, $BUFSIZ)) > 0) {
		    $message .= $buffer;
		    print DEBUG "ERROR ROFD read $nread\n" if $debug > 3;
			print DEBUG "ERROR ROFD read ".length($message)."\n"
				if $debug > 2;
			$bufferhex = unpack("H*", $message);
			print DEBUG "RECVHEX stdout ".$bufferhex."\n"
				if $printdatahex && $debug;
			print STDOUT "RECVHEX stdout ".$bufferhex."\n"
				if $printdatahex;
			print DEBUG "RECV stdout ".$buffer."\n"
				if $printdata && $debug;
			print STDOUT "RECV stdout ".$buffer."\n"
				if $printdata;
		} elsif ($nread == 0) {
		    print DEBUG "ERROR ROFD eof\n" if $debug;
		    @fids = handle_eof(fileno(ROFD));
		}
	}
	if (vec($rout, fileno(REFD), 1)) {
		print DEBUG "ERROR REFD ready\n" if $debug > 2;
		$buffer = '';
		$message = '';
		if (($nread = sysread(REFD, $buffer, $BUFSIZ)) > 0) {
		    $message .= $buffer;
			print DEBUG "ERROR REFD read $nread\n" if $debug > 3;
			# my (@lines, $l);
			# @lines = split('\n', $message);
			# foreach $l (@lines) {
			# 	print DEBUG "ERROR ".$l."\n";
			# }
			print THESTDERR $message;
			print DEBUG "ERROR REFD read ".length($message)."\n"
				if $debug > 2;
		} elsif ($nread == 0) {
		    print DEBUG "ERROR REFD eof\n" if $debug;
		    @fids = handle_eof(fileno(REFD));
		}
	}
}


# some support routines for the select code.
sub printVec {
    my($v) = @_;
    my($i, $result);

    for ($i = (8*length($v)) - 1; $i >= 0; $i--) {
	$result .= (vec($v, $i, 1)) ? "1" : "0";
    }
    $result;
}

sub fhbits {
	my(@fhlist) = @_;
	my($bits);
	for (@fhlist) {
		vec($bits, fileno($_), 1) = 1;
	}
	return $bits;
}
sub fdbits {
	my(@fdlist) = @_;
	my($bits);
	for (@fdlist) {
		vec($bits, $_, 1) = 1;
	}
	return $bits;
}

sub handle_eof {
	my($fid) = @_;
	my($e,@f);
	print DEBUG "HANDLE_EOF".@fids. "\n" if $debug;	
	foreach $e (@fids) {
		if ($fid != $e) {
			print DEBUG "HANDLE_EOF keeping ".$e. "\n" if $debug;
			push(@f, $e);
		} else {
			print DEBUG "HANDLE_EOF removing ".$e. "\n" if $debug;
		}
	}
	if (@f == 0) {
		print DEBUG "HANDLE_EOF fcount=0\n" if $debug;
		waitpid($cpid,0);
		my($stat) = $?;
		print DEBUG "ERROR waitpid status ($stat)\n" if $debug;
		exit(0);
	} else {
		print DEBUG "HANDLE_EOF fcount not 0\n" if $debug;
		@fids = @f;
		$rin = fdbits(@fids);
		$ein = $rin | $win;
	}
	return @fids;
}

sub handler {       # 1st argument is signal name
	my($sig) = @_;
	print DEBUG "ERROR Caught a SIG$sig\n" if $debug;
}
