/*
 * set.c
 *
 * Author: Rene G. de Vries
 * Date  : Thu Nov 15 18:48:17 CET 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "set.h"

#define slotsize (sizeof(TBarray)*8)


typedef unsigned int TBarray;

typedef struct sSettype {
  int nrslots;
  int nrbits;         
} sSettype;

typedef struct sSet {
  Settype type;
  TBarray *barray;
} sSet;


Settype
DeclaresetType(int max_cardinality)
{ Settype t = (Settype) malloc (sizeof(struct sSettype));

  t->nrbits=max_cardinality;
  t->nrslots = (max_cardinality/slotsize) + ((max_cardinality%(slotsize))?1:0);
  return t;
}

Settype
Typeofset(Set s)
{ return s->type;
}


void
Insertset(Set s, int m)
{
  if(m >= s->type->nrbits) {
  	fprintf(stderr, "set error (Insertset: element index (%d) >= cardinality (%d))\n",
  	       m, s->type->nrbits);
    exit(1);
  }
  s->barray[m/slotsize] |= 1 << (m%slotsize);
}

void
Removeset(Set s, int m)
{
  if(m >= s->type->nrbits) {
  	fprintf(stderr, "set error (Removeset: element index (%d) >= cardinality (%d))\n",
  	       m, s->type->nrbits);
    exit(1);
  }
  s->barray[m/slotsize] &= ~(1 << (m%slotsize));
}

char
Isinset(Set s, int m)
{
  if(m >= s->type->nrbits) {
  	fprintf(stderr, "set error (Isinset: element index (%d) >= cardinality (%d))\n",
  	       m, s->type->nrbits);
    exit(1);
  }
  return (s->barray[m/slotsize] & 1 << (m%slotsize))!= 0;
}

char
Isemptyset(Set s)
{
  int i;
  for (i=0;(i<s->type->nrslots-1) && (s->barray[i] == 0);i++);
  return (s->barray[i] == 0);
}

void
Unionset(Set sd, Set s1, Set s2)
{
  int i;

  if ((s1->type != s2->type) || (s2->type != sd->type)) {
    fprintf(stderr, "set error (union: type differs)\n");
    exit(1);
  }
  for (i=0;i<s1->type->nrslots;i++)
    sd->barray[i] = s1->barray[i] | s2->barray[i];
}

void
Copyset(Set sd, Set s)
{
  int i;

  if (sd->type != s->type ) {
    fprintf(stderr, "set error (copy: type differs)\n");
    exit(1);
  }
  for (i=0;i<s->type->nrslots;i++)
    sd->barray[i] = s->barray[i];
}

void
Intersectionset(Set sd, Set s1, Set s2)
{
  int i;

  if ((s1->type != s2->type) || (s2->type != sd->type)) {
    fprintf(stderr, "set error (intersection: type differs)\n");
    exit(1);
  }
  for (i=0;i<s1->type->nrslots;i++)
    sd->barray[i] = s1->barray[i] & s2->barray[i];
}

void
Diffset(Set sd, Set s1, Set s2)
{
  int i;

  if ((s1->type != s2->type) || (s2->type != sd->type)) {
    fprintf(stderr, "set error (diff: type differs)\n");
    exit(1);
  }
  for (i=0;i<s1->type->nrslots;i++)
    sd->barray[i] = s1->barray[i] & ~(s2->barray[i]);
}

Set
Newset(Settype t)
{  int i;
 
   Set s; 
   s = (sSet *) malloc (sizeof(sSet));
   s->type = t;
   s->barray = (TBarray *) malloc(sizeof(TBarray)*s->type->nrslots);
   for (i=0;i<(s->type->nrslots);i++) s->barray[i]=0;
   return s;
}

void
Emptyset(Set s)
{  int i;
   for (i=0;i<(s->type->nrslots);i++) s->barray[i]=0;
   return; 
}

void
Delset(Set s)
{ 
  free(s->barray);
  free(s);
}
 
void
Printset(FILE *fd, Set s)
{ int i;
  TBarray h=1;

  for (i=0;i<(s->type->nrbits);i++) {
    if ((s->barray[i/slotsize]) & h) fprintf (fd, "%d ",i);
    h = h << 1;if (!h) h=1;
  }
}

void
Printsets(FILE *fd, void (*ptr)(FILE *, int),  Set s, char *sep)
{ int i;
  int j=0;
  TBarray h=1;

  for (i=0;i<(s->type->nrbits);i++) {
    if ((s->barray[i/slotsize]) & h) {
      if(j) fprintf(fd, "%s", sep);
      j=1; /* nice sep-seperated list */
      (*ptr)(fd,i);
    }
    h = h << 1;if (!h) h=1;
  }
}

void
setforeach(void (*fn)(int, void*), void*p, Set s)
{ int i;
  TBarray h=1;

  for (i=0;i<(s->type->nrbits);i++) {
    if ((s->barray[i/slotsize]) & h) {
      (*fn)(i, p);
    }
    h = h << 1;
    if (!h) h=1;
  }
}

void
settest()
{ 
  Settype t = DeclaresetType(100);
  Set s1,s2,sd;

  s1=Newset(t);

  s2=Newset(t);

  sd=Newset(t);

  
  Insertset(s1,99);
  Insertset(s2,17);
  Insertset(s2,99);
  Printset(stdout,s1);
  Printset(stdout,s2);
 
  Unionset(sd,s1,s2);
  Printset(stdout,sd);
  Intersectionset(sd,s1,s2);
  Printset(stdout,sd);
  Diffset(sd,s2,s1);
  Printset(stdout,sd);
  printf("inset: %d,%d\n",Isinset(s1,34),Isinset(s1,99));
  printf("empty: %d\n",Isemptyset(s1));
  Removeset(s1,99);
  printf("empty: %d\n",Isemptyset(s1));
  Delset(sd);
  Delset(s1);
  Delset(s2);
}
