#ifndef SPEX_DOMAIN
#define SPEX_DOMAIN

#include "spexSet.h"
#include "spexRange.h"

/*
Common tasks for Domains:

Create a new domain (from scratch)

Compare two domains (equal)

Create a new domain by:
- Shifting a domain by a value
- Removing elements from a domain greater or smaller than a value
- Intersecting a domain with another domain
- Union of a domain with another domain
*/

#ifndef DOMAIN
#define DOMAIN	unsigned char
#endif
typedef DOMAIN Domain;

#ifndef STORAGE_SIZE
#define STORAGE_SIZE 254
#endif

// Creates a new domain from the string specified. This string has to be in the
// form of "(el), (el)", where (el) can be either "(val)" or "(val)..(val)", where
// (val) is an integer. 
Domain newDomain(const char * ranges);
Domain newSingleDomain(RANGE_VALUE_TYPE lb, RANGE_VALUE_TYPE ub);

// Returns a pointer to buffer containing a string representation of the domain.
char * 	DomainToString(Domain);

// Internal functions for storing and loading Sets
Domain DStoreSet(DSetP);
DSetP  DGetSet(Domain);


// Check if domain 'a' is equal to 'b'.
char	DomainEqual(Domain a, Domain b);

// Move the domain up- or downwards.
Domain	DomainMove(Domain, RANGE_VALUE_TYPE);

Domain  DomainAddOne(Domain a, RANGE_VALUE_TYPE lb, RANGE_VALUE_TYPE ub);

char	DHasGt(Domain d, RANGE_VALUE_TYPE value);
char	DHasGtEq(Domain d, RANGE_VALUE_TYPE value);
char	DHasSt(Domain d, RANGE_VALUE_TYPE value);
char	DHasStEq(Domain d, RANGE_VALUE_TYPE value);
char	DHas(Domain d, RANGE_VALUE_TYPE value);



// Removes all values from the domain smaller or equal to 'value'. 
// (Domain > value)
Domain	DomainGt(Domain d, RANGE_VALUE_TYPE value);

// Removes all values from the domain smaller than 'value'. 
// (Domain >= value)
Domain	DomainGtEq(Domain d, RANGE_VALUE_TYPE value);

// Removes all values from the domain greater or equal to 'value'. 
// (Domain < value)
Domain	DomainSt(Domain d, RANGE_VALUE_TYPE value);

// Removes all values from the domain greater than 'value'. 
// (Domain <= value)
Domain	DomainStEq(Domain d, RANGE_VALUE_TYPE value);

// Removes all values from the domain not equal to 'value'.
Domain	DomainEq(Domain d, RANGE_VALUE_TYPE value);

// Removes all values from the domain equal to 'value'.
Domain	DomainNEq(Domain d, RANGE_VALUE_TYPE value);


Domain	DomainGt2(Domain *d, RANGE_VALUE_TYPE value);
Domain	DomainSt2(Domain *d, RANGE_VALUE_TYPE value);


// Removes all values not between 'lb' and 'ub' (resp. lower- and upper bound)
Domain	DomainBetween(RANGE_VALUE_TYPE lb, Domain d, RANGE_VALUE_TYPE ub);


// Removes all values from 'a' which are not in 'b'.
Domain	DomainIntersect(Domain a, Domain b);

// Combines all values from 'a' and 'b'.
Domain	DomainUnion(Domain, Domain);

// Removes all values from 'a' which are in 'b'.
Domain	DomainRemove(Domain a, Domain b);


// Not sure if this will be necessary:
// Domain	DomainAdd(Domain a, Domain b)
// The result is a Domain where every combination of 
// (pick(a) + pick(b)) is represented

#endif
