#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "spexIO.h"
#include "spex_mtab.h"
#include "spexDomain.h"


char *trans_text;
Var_Store *trans_vars;
Var_Store *last_trans_var;

#if !defined(MAX_TRANS_LENGTH) || MAX_TRANS_LENGTH < 128
#define MAX_TRANS_LENGTH 128
#endif

static void
reserve(char **s, char **p, char **e, int n)
{
	int used, len;

	used = *p - *s;
	len = *e - *s;

//fprintf(stderr, "A_DEBUG reserve b s=%x p=%x e=%x n=%d used=%d len=%d\n", *s, *p, *e, n, used, len);
	if (used + n >= len ) {
		len += (n < 10) ? 10 : n;
		if (*s == 0)
			*s = malloc(len);
		else
			*s = realloc(*s, len);
		if (*s == 0) {
			fprintf(stdout, "A_ERROR Could not allocate space for transition text.\n");
			fflush(stdout);
			exit(1);
		}
		*e = *s + len; 
		*p = *s + used;
	}
//fprintf(stderr, "A_DEBUG reserve e s=%x p=%x e=%x n=%d used=%d len=%d\n", *s, *p, *e, n, used, len);
}


// A call to spex__io will look something like:
// spex__io(format, varfields_1, varpointer_1, varranges_1, vardesc_2, varpointer_2, varranges_2...)
// where varfields_1 consists of one or more fields.
// for 'in' parameters it consists of: vardesc, varpointer, varranges
// for 'out' parameters it consists of: value
// where the 'format' string contains a number of '%' conversions, similar to scanf.
// The spex__io function will create Var_Stores for each var, containing normalized and
// original variable names, the type (as a string) and a pointer to the location of the variable
// These Var_Stores will be added in a global linked list to be used by the SPEX program.
void spex__io(char *str, ...) {
	char *s, *e;
	char *dst, *d, *edst;
	char *var_str;
	void *var_p;
	char *var_ranges;
	Var_Store *vstore;
	int int_index, short_index, byte_index, mtype_index;
	char tmp_line[80];
	va_list arg_list;
	
	int_index = 1;
	short_index = 1;
	byte_index = 1;
	mtype_index = 1;


	s = str;
	e = s + strlen(s);

	dst = 0;
	d = dst;
	edst = dst;

	reserve(&dst, &d, &edst, 1);
	*d = '\0';
	
	//fprintf(stderr, "A_DEBUG (input) trans_text set to: %s\n", trans_text);	
	
	va_start(arg_list, str);
	
	for (;s < e; s++) {
		if ( *s == '\\' ) {
			s++;
			if (d + 1 >= edst)
				reserve(&dst, &d, &edst, 1);
			*d = *s;
			d++;
		} else if ( *s == '%' && (s + 1) < e) {
			s++;
			// fprintf(stderr, "A_DEBUG found a %% in input format. (type:%c)\n", *s);
			if (*s == '%') {
				// skip both '%' characters.
				if (d + 2 >= edst)
					reserve(&dst, &d, &edst, 2);
				*d = '%';
				d++;
				*d = *s;
				d++;
			} else if (*s == '&') {
				s++;
				// Add one pointer to the list
				var_str = va_arg(arg_list, char *);
				var_p = va_arg(arg_list, void *);
				var_ranges = va_arg(arg_list, void *);

				vstore = new_var_store();
				append_var_store(vstore);

				if (d + 2 >= edst)
					reserve(&dst, &d, &edst, 2);
				*d = '%';
				d++;
				*d = *s;
				d++;

	/*
Typename	C-equivalent	Macro in limits.h	Typical Range
bit or bool	bit-field	-			0..1
byte		uchar		CHAR_BIT width in bits  0..255
short		short		SHRT_MIN..SHRT_MAX	-2^15 - 1 .. 2^15 - 1
int		int		INT_MIN..INT_MAX	-2^31 - 1 .. 2^31 - 1
	*/

				
				switch (*s) {
					case 'd' : {
						sprintf(tmp_line, "var_int$%d%c", int_index, '\0');
						vstore->type = strdup("int");
						vstore->pointer.i = var_p;
						int_index++;
						break;
					} 
					case 'c': {
						sprintf(tmp_line, "var_byte$%d%c", byte_index, '\0');
						vstore->type = strdup("byte");
						vstore->pointer.c = var_p;
						byte_index++;
						break;
					}
					case 'h' : {
						sprintf(tmp_line, "var_short$%d%c", short_index, '\0');
						vstore->type = strdup("short");
						vstore->pointer.s = var_p;
						short_index++;
						break;
					}
					case 'e' : {
						sprintf(tmp_line, "var_mtype$%d%c", mtype_index, '\0');
						vstore->type = strdup("mtype");
						vstore->pointer.c = var_p;
						mtype_index++;
						break;
					}
					default : {
						printf("A_ERROR Unknown conversion character '%c' in label.\n", *s);
						exit(1);
					}
				}					
				vstore->norm_var_name = strdup(tmp_line);
				vstore->org_var_name = strdup(var_str);
				vstore->var_domain = newDomain(var_ranges);
				
			} else {
				// output
				// Promela (apperantly) doesn't support proper printing of shorts, so we use our own.
				// All va_arg() read ints, because optional parameters are promoted to these.
				switch (*s) {
					case 'd' : {
						int i = va_arg(arg_list, int);
						if (d + 15 >= edst)
							reserve(&dst, &d, &edst, 15);
						sprintf(d, "%d", i);
						d += strlen(d);
						break;
					} 
					case 'c': {
						int i  = (unsigned char) va_arg(arg_list, int);
						if (d + 15 >= edst)
							reserve(&dst, &d, &edst, 15);
						sprintf(d, "%d", i);
						d += strlen(d);
						break;
					}
					case 'h' : {
						int i = (short) va_arg(arg_list, int);
						if (d + 15 >= edst)
							reserve(&dst, &d, &edst, 15);
						sprintf(d, "%d", i);
						d += strlen(d);
						break;
					}
					case 'e' : {
						int i = (unsigned char) va_arg(arg_list, int);
						if (d + 30 >= edst)
							reserve(&dst, &d, &edst, 30);
						sprintf(d, "%s", mstring(i));
						d += strlen(d);
						break;
					}
					case 's' : {
						char *s;
						int i;
						
						s = (char *) va_arg(arg_list, char *);
						if ( s == NULL ) {
							s = "(null)";
						} 
						i = strlen(s);
						if ( d + i >= edst)
							reserve(&dst, &d, &edst, i);
						sprintf(d, "%s", s);
						d += i;
						break;
					}
							
					default : {
						printf("A_ERROR Unknown conversion character '%c' in label.\n", *s);
						exit(1);
					}
				}					
			}
		} else {
			if ( *s == '%')
				fprintf(stderr, "A_WARNING '%%' at end of spex__inputs format string: %s\n", str);	
			if (d + 1 >= edst)
				reserve(&dst, &d, &edst, 1);
			*d = *s;
			d++;
		}
	}
	reserve(&dst, &d, &edst, 1);
	*d = '\0';
	d++;
	trans_text = dst;
	fprintf(stderr, "A_DEBUG (spexio) trans_text set to: %s\n", trans_text);	
	
	//p = va_arg(arg_list, int *);
		
	//printf("%d", *p);
	
	va_end(arg_list);
	
}

// A call to spex__inputs will look something like:
// spex__inputs(format, vardesc_1, varpointer_1, varranges_1, vardesc_2, varpointer_2, varranges_2...)
// where the 'format' string contains a number of '%' conversions, similar to scanf.
// The spex__inputs function will create Var_Stores for each var, containing normalized and
// original variable names, the type (as a string) and a pointer to the location of the variable
// These Var_Stores will be added in a global linked list to be used by the SPEX program.
void spex__inputs(char *str, ...) {
	char *s, *e;
	char *var_str;
	void *var_p;
	char *var_ranges;
	Var_Store *vstore;
	int int_index, short_index, byte_index, mtype_index;
	char tmp_line[80];
	va_list arg_list;
	
	int_index = 1;
	short_index = 1;
	byte_index = 1;
	mtype_index = 1;


	s = str;
	e = s + strlen(s);
	
	trans_text = strdup(str);
	//fprintf(stderr, "A_DEBUG (input) trans_text set to: %s\n", trans_text);	
	
	va_start(arg_list, str);
	
	for (;s < e; s++) {
		if ( *s == '\\' ) {
			s++;
		} else if ( *s == '%' && (s + 1) < e) {
			s++;
			//fprintf(stderr, "A_DEBUG found a %% in input format. (type:%c)\n", *s);
			if (*s == '%') {
				// skip both '%' characters.
			} else {
				// Add one pointer to the list
				var_str = va_arg(arg_list, char *);
				var_p = va_arg(arg_list, void *);
				var_ranges = va_arg(arg_list, void *);

				vstore = new_var_store();
				append_var_store(vstore);

	/*
Typename	C-equivalent	Macro in limits.h	Typical Range
bit or bool	bit-field	-			0..1
byte		uchar		CHAR_BIT width in bits  0..255
short		short		SHRT_MIN..SHRT_MAX	-2^15 - 1 .. 2^15 - 1
int		int		INT_MIN..INT_MAX	-2^31 - 1 .. 2^31 - 1
	*/

				// TODO Actually, we should check here if a variable appears multiple times in one place.
				// E.g. from_env ! var_int$1 ! var_int$2 ! var_int$1
				switch (*s) {
					case 'd' : {
						sprintf(tmp_line, "var_int$%d%c", int_index, '\0');
						vstore->type = strdup("int");
						vstore->pointer.i = var_p;
						int_index++;
						break;
					} 
					case 'c': {
						sprintf(tmp_line, "var_byte$%d%c", byte_index, '\0');
						vstore->type = strdup("byte");
						vstore->pointer.c = var_p;
						byte_index++;
						break;
					}
					case 'h' : {
						sprintf(tmp_line, "var_short$%d%c", short_index, '\0');
						vstore->type = strdup("short");
						vstore->pointer.s = var_p;
						short_index++;
						break;
					}
					case 'e' : {
						sprintf(tmp_line, "var_mtype$%d%c", mtype_index, '\0');
						vstore->type = strdup("mtype");
						vstore->pointer.c = var_p;
						mtype_index++;
						break;
					}
					default : {
						printf("A_ERROR Unknown conversion character '%c' in label.\n", *s);
						exit(1);
					}
				}					
				vstore->norm_var_name = strdup(tmp_line);
				vstore->org_var_name = strdup(var_str);
				vstore->var_domain = newDomain(var_ranges);
				
			}
		} else if ( *s == '%') {
			fprintf(stderr, "A_WARNING '%%' at end of spex__inputs format string: %s\n", str);	
		}
	}
	
	//p = va_arg(arg_list, int *);
		
	//printf("%d", *p);
	
	va_end(arg_list);
	
}

// Print arguments in the format described by 'trans' (analogous to printf) to the 'trans_text' variable,
// that points to some newly allocated space. The user is responsible for freeing the allocated space after
// being done with it.
void spex__outputs(char *trans, ...) {
	char *s, *e;
	char *dst, *d, *edst;
	//int ndst;
	va_list arg_list;
	
	s = trans;
	e = s + strlen(s);

	dst = 0;
	d = dst;
	edst = dst;

	reserve(&dst, &d, &edst, 1);
	*d = '\0';
	
	va_start(arg_list, trans);
	
	for (;s < e; s++) {
		if ( *s == '\\'  && (s + 1) < e) {
			s++;
			if (d + 1 >= edst)
				reserve(&dst, &d, &edst, 1);
			*d = *s;
			d++;
		} else if ( *s == '%' && (s + 1) < e) {
			s++;
			//fprintf(stderr, "A_DEBUG found a %% in output format. (type:%c)\n", *s);
			if (*s == '%') {
				if (d + 1 >= edst)
					reserve(&dst, &d, &edst, 1);
				*d = *s;
				d++;
			} else {
				// Promela (apperantly) doesn't support proper printing of shorts, so we use our own.
				// All va_arg() read ints, because optional parameters are promoted to these.
				switch (*s) {
					case 'd' : {
						int i = va_arg(arg_list, int);
						if (d + 15 >= edst)
							reserve(&dst, &d, &edst, 15);
						sprintf(d, "%d", i);
						d += strlen(d);
						break;
					} 
					case 'c': {
						int i  = (unsigned char) va_arg(arg_list, int);
						if (d + 15 >= edst)
							reserve(&dst, &d, &edst, 15);
						sprintf(d, "%d", i);
						d += strlen(d);
						break;
					}
					case 'h' : {
						int i = (short) va_arg(arg_list, int);
						if (d + 15 >= edst)
							reserve(&dst, &d, &edst, 15);
						sprintf(d, "%d", i);
						d += strlen(d);
						break;
					}
					case 'e' : {
						int i = (unsigned char) va_arg(arg_list, int);
						if (d + 30 >= edst)
							reserve(&dst, &d, &edst, 30);
						sprintf(d, "%s", mstring(i));
						d += strlen(d);
						break;
					}
					case 's' : {
						char *s;
						int i;
						
						s = (char *) va_arg(arg_list, char *);
						if ( s == NULL ) {
							s = "(null)";
						} 
						i = strlen(s);
						if ( d + i >= edst)
							reserve(&dst, &d, &edst, i);
						sprintf(d, "%s", s);
						d += i;
						break;
					}
							
					default : {
						printf("A_ERROR Unknown conversion character '%c' in label.\n", *s);
						exit(1);
					}
				}					
			}
		} else {
			if (d + 1 >= edst)
				reserve(&dst, &d, &edst, 1);
			*d = *s;
			d++;
		}
	}
	reserve(&dst, &d, &edst, 1);
	*d = '\0';
	d++;
	trans_text = dst;
	
	//fprintf(stderr, "A_DEBUG (output) trans_text set to: %s\n", trans_text);
	va_end(arg_list);	
}

Var_Store *new_var_store() {
	Var_Store *v;
	
	v = calloc(1, sizeof(Var_Store));
	if (v == NULL) {
		printf("A_ERROR Unable to allocate new variable storage. Program will exit now.\n");
		exit(-1);
	}
	return v;	
}

void free_var_store(Var_Store *v) {
	free(v->norm_var_name);
	free(v->org_var_name);
	free(v->type);
	free(v);
}

void free_var_stores(Var_Store **stack) {
	Var_Store *curp;
	Var_Store *curp_next;
	
	// Remove all indices that were stored for callback.
	for (curp = *stack; curp != NULL; curp = curp_next) {
		curp_next = curp->next;
		free_var_store(curp);
	}
	*stack = NULL;	
}

Var_Store *append_var_store(Var_Store *vstore) {
	if ( trans_vars == NULL ) {
		trans_vars = last_trans_var = vstore;
	} else {
		last_trans_var->next = vstore;
		last_trans_var = vstore;
	}
	
	return vstore;	
}

Var_Store *dup_var_store(Var_Store *vstore) {
	Var_Store *new_vstore;
	if (vstore == NULL) {
		return NULL;
	} else {
		new_vstore = new_var_store();
		new_vstore->norm_var_name = strdup(vstore->norm_var_name);
		new_vstore->org_var_name = strdup(vstore->org_var_name);
		new_vstore->type = strdup(vstore->type);
		new_vstore->pointer = vstore->pointer;
		new_vstore->var_domain = vstore->var_domain;
		
		return new_vstore;
	}
}

