#
# cadp.pm - mkprimer module for the CADP LOTOS primer
# will be ``required'' by the mkprimer program
#
# $Id: cadp.pm,v 1.5.2.5 2004/06/11 21:52:48 belinfan Exp $
# <Axel.Belinfante@cs.utwente.nl>, 1999
#
#
# ======================================
package cadp;

$libname = 'libcadp';
$gcname = 'gc';

if ( -r "$libdir/lib".$gcname.".a") {
  $gc_flags = "-L$libdir -l$gcname";
} else {
  $gc_flags = "";
}

%suffix = (
			".lotos"   =>   "LOTOS",
			".lot"     =>   "LOTOS",
			".aut"     =>   "AUT",
			".bcg"     =>   "BCG",
			".fc2"     =>   "FC2",
	     );

%lang = (
	"LOTOS" =>
	 sub  {
		my($specfile, $target, $prefix, $libdir,
		   $specpath, $specname, $specsuffix, $exesuffix) = @_;
		my(@cmds) = (
			"caesar.open $specfile $gc_flags $libdir/$libname.a < /dev/null",
			"mv $libname$exesuffix $target$exesuffix",
		          );
		::Exec(@cmds);
	},
	"BCG" =>
	 sub  {
		my($specfile, $target, $prefix, $libdir,
		   $specpath, $specname, $specsuffix, $exesuffix) = @_;
		my(@cmds) = (
			"bcg_open $specfile $gc_flags $libdir/$libname.a < /dev/null",
			"mv $libname$exesuffix $target$exesuffix",
		          );
		::Exec(@cmds);
	},
	"AUT-CADP" =>
	 sub  {
		my($specfile, $target, $prefix, $libdir,
		   $specpath, $specname, $specsuffix, $exesuffix) = @_;
		my(@cmds) = (
			"bcg_io -aldebaran - -bcg $specpath$specname.bcg < $specfile",
			"bcg_open $specpath$specname.bcg $gc_flags $libdir/$libname.a < /dev/null",
			"mv $libname$exesuffix $target$exesuffix",
		          );
		::Exec(@cmds);
	},
	"FC2" =>
	 sub  {
		my($specfile, $target, $prefix, $libdir,
		   $specpath, $specname, $specsuffix, $exesuffix) = @_;
		my(@cmds) = (
			"bcg_io -fc2 - -bcg $specpath$specname.bcg < $specfile",
			"bcg_open $specpath$specname.bcg $gc_flags $libdir/$libname.a < /dev/null",
			"mv $libname$exesuffix $target$exesuffix",
		          );
		::Exec(@cmds);
	},
);

::Register(\%lang, \%suffix);

1;
