# =========================================================================
# $Id: main_util.tcl,v 1.1.2.37 2008/04/11 11:54:05 belinfan Exp $
# =========================================================================

proc primer_init {file algo ioct complete} {
	global susp_history
	global primer
	# debugMsg "primer_init ($file)($algo)($ioct)($complete)"
	set susp_history(input) {}
	set susp_history(output) {}
	set primer(algorithm) $algo
	set primer(ioct) $ioct
	event_kind_init $file
	foreach kind [split $complete ,] {
		debugMsg "primer_init: $kind"
		switch -- $kind {
			  input -
			  output {
				set primer(complete,$kind) 1
			} default {
				warningMsg "Unknown iokind for completion: $kind"
			}
		}
	}
	reset_menu_verdict
}

proc primer_start {} {
	explorer_init
	set initial [explorer_start]
	set initial_state [lindex $initial 0]
	set preds_solved  [lindex $initial 1]
	set preds  [lindex $initial 2]
	set freevars  [lindex $initial 3]
	set men_tra_ign_id [obs_menu $initial_state]
	set transitive [lindex $men_tra_ign_id 1]
	set ignorable [lindex $men_tra_ign_id 2]
	set identical [lindex $men_tra_ign_id 3]
	if {! $preds_solved} {
		warningMsg "Predicates Unsolved for <root> $initial_state: $preds"
	}
	primer_set_super $initial_state $transitive $ignorable $identical
	primer_store_super
}

proc primer_menu {prefix keyword kinds chan_sug} {
	set chan_opt {}
	if {[string compare $chan_sug ""] != 0} {
		if {![is_channel $chan_sug]} {
			puts "A_ERROR WrongValue: not a channel: \"$chan_sug\""
			return
		} elseif {[lsearch $kinds [get_kind $chan_sug]] < 0} {
			puts "A_ERROR Inconsistency: wrong iokind (not $keyword) channel: \"$chan_sug\""
			return
		} else {
			lappend chan_opt $chan_sug
		}
	}
	puts "A_[set keyword]_START"
	foreach el [mr_list $kinds $chan_opt] {
		set kind [lindex $el 0]
		set chan [lindex $el 1]
		set ev [lindex $el 2]
		set rest [lindex $el 3]
		set etuple [lindex $el 4]
		set preds [lindex $etuple 4]
		set attrs [lindex $el 5]
		switch -- $kind {
			  inputsuspension -
			  outputsuspension {
				set s_attr "\tsuspension=1"
			} input -
			  output {
				set s_attr "\tsuspension=0"
			} default {
				set s_attr ""
			}
		}
		set c_attr "channel=$chan"
		if {[string compare $preds ""] == 0} {
				set p_attr ""
		} else {
				set p_attr "\tpredicates=$preds"
		}
		set i_str [join $rest ,]
		set astr ""
		foreach a $attrs {
			set n [lindex $a 0]
			set v [lindex $a 1]
			append astr "\t$n=$v"
		}
		puts "$prefix event=$ev$s_attr\t$c_attr$p_attr\tinit=$i_str$astr"
	}
	puts "A_[set keyword]_END"
}

proc primer_parse {line} {
	# debugMsg "ToParse: ($line)"
	if {[regexp {^[ \t]*([cC]_[a-zA-Z0-9_]+)([ \t].*)?} $line match cmd rest]} {
		set the_rest [string trim $rest]
		return [list 1 [list $cmd $the_rest]]
	} else {
		return [list 0]
	}
}

proc primer_get_input {ev chan s preds} {
	global primer
	if {([string compare $primer(algorithm) sim] == 0) ||
	    ([string compare $ev ""] != 0) ||
	    ([string compare $s ""] != 0)} {
		return [primer_check_next 0 GETINPUT input $ev $chan $s $preds]
	} else {
		return [primer_random_next 0 GETINPUT input $ev $chan $s $preds]
	}
}
proc primer_do_input {ev chan s preds} {
	global primer
	debugMsg "primer_do_input: ($ev)($chan)($s)"
#	if {([string compare $primer(algorithm) sim] == 0) ||
#	    ([string compare $ev ""] != 0) ||
#	    ([string compare $s ""] != 0)} {
		return [primer_check_next 1 INPUT input $ev $chan $s $preds]
#	} else {
#		return [primer_random_next 1 INPUT input $ev $chan $s $preds]
#	}
}
proc primer_get_output {ev chan s preds} {
	global primer
	if {([string compare $primer(algorithm) sim] != 0) ||
	    ([string compare $ev ""] != 0) ||
	    ([string compare $s ""] != 0)} {
		return [primer_check_next 0 GETOUTPUT output $ev $chan $s $preds]
	} else {
		return [primer_random_next 0 GETOUTPUT output $ev $chan $s $preds]
	}
}
proc primer_do_output {ev chan s preds} {
	global primer
	if {([string compare $primer(algorithm) sim] != 0) ||
	    ([string compare $ev ""] != 0) ||
	    ([string compare $s ""] != 0)} {
		return [primer_check_next 1 OUTPUT output $ev $chan $s $preds]
	} else {
		return [primer_random_next 1 OUTPUT output $ev $chan $s $preds]
	}
}

proc primer_check_next {move keyword iokind ev chan s preds} {
	set comb [primer_combine_fields $iokind $ev $chan $s C_[set keyword]]
	debugMsg "primer_check_next $move $keyword $iokind $chan ($ev) ($preds): ($comb)"
	if {[lindex $comb 0]} {
		set longkind [lindex $comb 1]
		set ev [lindex $comb 2]
		set chan [lindex $comb 3]
		set s [lindex $comb 4]
		primer_do_next $move $keyword $iokind $ev $chan $s $preds
	} else {
		puts [lindex $comb 1]
	}
}
proc primer_random_next {move keyword iokind ev chan s preds} {
	set comb [primer_choose_next $iokind $ev $chan $s C_[set keyword]]
	debugMsg "primer_random_next $move $keyword $iokind $chan ($ev) ($preds): ($comb)"
	if {[lindex $comb 0]} {
		set longkind [lindex $comb 1]
		set ev [lindex $comb 2]
		set chan [lindex $comb 3]
		set s [lindex $comb 4]
		primer_do_next $move $keyword $iokind $ev $chan $s $preds
	} else {
		puts [lindex $comb 1]
	}
}

# very rough string comparison
proc pred_compare {p1 p2} {
	regsub -all { } $p1 {} r_p1
	regsub -all { } $p2 {} r_p2
	# cannot use -nocase flag with older tcl versions
	if {[string compare [string tolower $r_p1] [string tolower $r_p2]] == 0} {
		return 1
	} else {
		return 0
	}
}
proc primer_do_next {move keyword iokind ev chan s preds} {
	global susp_history
	global primer
	debugMsg "primer_do_next $move $keyword $iokind $chan ($ev) ($preds)"
	if {([string compare $s ""] != 0) && $s} {
		set trying_suspension $chan
		set evs_newS [primer_next [set iokind]suspension $ev $chan $preds]
	} else {
		set trying_suspension ""
		set evs_newS [primer_next [set iokind] $ev $chan $preds]
	}
	set evs [lindex $evs_newS 0]
	set newS [lindex $evs_newS 1]
	set errmsg [lindex $evs_newS 2]
	debugMsg "primer_do_next ($evs)($preds)"
	set evs_l [llength $evs]
	if {$evs_l > 0} {
		set res_evs {}
		set res_newS {}
		foreach ev $evs {
			set p_ev [lindex $ev 4]
			if {[pred_compare $p_ev $preds]} {
				lappend res_evs $ev
				lappend res_newS [lindex $ev 0]
				debugMsg "	primer_do_next $move $keyword $iokind $chan ($ev) ($preds): multiple event, matches preds: ($p_ev)($preds)"
			} else {
				debugMsg "	primer_do_next $move $keyword $iokind $chan ($ev) ($preds): multiple event, not match preds: ($p_ev)($preds)"
			}
		}
		set res_l [llength $res_evs]
		if {$res_l == 0 || $res_l == $evs_l} {
			set res_ev [lindex $evs 0] ;# just take the first one, for the moment
			debugMsg "	primer_do_next multiple event res_l ($res_l) == 0 || res_l == evs_l ($evs_l)"
			debugMsg "	primer_do_next multiple event: using $res_ev"
		} else {
			set res_ev [lindex $res_evs 0] ;# just take the first one, for the moment
			set newS $res_newS
			debugMsg "	primer_do_next multiple event res_l NOT(($res_l) == 0 || res_l == evs_l ($evs_l))"
			debugMsg "	primer_do_next multiple event: using $res_ev"
		}
		set pr_ev [lindex $res_ev 3]
		set p_ev [lindex $res_ev 4]
		if {$move} {
			set canonical1 [expansion_canonicalize $newS]
			set seq1 [primer_mksupersequence $canonical1]
			set idnr [primer_get_init_id $seq1]
			if {0 && $idnr != {}} {
				primer_retr_super $idnr
			} else {
				set men_tra_ign_id [obs_menu $newS]
				set transitive [lindex $men_tra_ign_id 1]
				set ignorable [lindex $men_tra_ign_id 2]
				set identical [lindex $men_tra_ign_id 3]
				primer_set_super $newS $transitive $ignorable $identical
				primer_store_super
			}
		}
	} else {
		# get rid of the edge info in the current state representation
		# or are there cases where we want it to be there??
		# One bug could be that we remove tau (internal) transition
		# edges that relate the states in the current superstate:
		# it might be nicer to highlight those edges.
		# Hmm, we try to fix the bug by not removing the edges
		# from the trans states in primer_strip_edge_super
		primer_strip_edge_super
		if {[info exists primer(complete,$iokind)] && $primer(complete,$iokind)} {
			set pr_ev $ev
		} else {
			puts "A_[set keyword]_ERROR reason=$errmsg"
			return
		}
	}
	if {[string compare $trying_suspension ""] != 0} {
		set s_attr "\tsuspension=1"
		if {[lsearch $susp_history($iokind) $trying_suspension] < 0} {
			if {$move} {
				lappend susp_history($iokind) $trying_suspension
			}
		}
	} else {
		set s_attr "\tsuspension=0"
		switch -- $iokind {
			input {
				if {$move} {
					set susp_history(output) {}
				}
			} output {
				if {$move} {
					set susp_history(input) {}
				}
			}
		}
	}
	if {[catch {event_kind $pr_ev} kind_chan]} {
		puts "A_ERROR UnknownIOKind:: event=$pr_ev"
		return
	}
	set c [lindex $kind_chan 1]
	set c_attr "channel=$c"
	if {[string compare $p_ev ""] == 0} {
		set p_attr ""
	} else {
		set p_attr "\tpredicates=$p_ev"
	}
	puts "A_[set keyword]_OK event=$pr_ev$s_attr\t$c_attr$p_attr"
}

# primer_combine_fields:
# combines the given fields; return either
#    [list 1 kind ev chan s] (succesfull combined fields)
#    [0 errmsg]              (unsufficient or incomplete info)
#
# Assumptions:
#  - we either generate or recognize suspension-events ((m)ioco resp traces)
#  - suspension-events may have an empty-string representation;
#    this should not pose problems, because we can describe them
#    using fields chan and s.
#  - each given field may be emtpy string; this only poses problems
#    if the other fields do not contain sufficient information
#    to have a result.
# NOTE: the returned kind will be one of:
#  input, output, inputsuspension, outputsuspension
proc primer_combine_fields { kind ev chan s cmd } {
	switch -- $kind {
		input -
		output {
		} default {
			return [list 0 "A_ERROR InternalError: primer_combine_fields: kind notin (input,output): \"$kind\""
		}
	}
	if {[string compare $chan ""] != 0} {
		if {[is_channel $chan]} {
			set channel $chan
		} else {
			return [list 0 "A_ERROR WrongValue: not a channel: \"$chan\""]
		}
	}
	if {[string compare $s ""] != 0} {
		if {$s} {
			set suspension 1
		} else {
			set suspension 0
		}
	}
	if {([string compare $ev ""] == 0)} {
		if {([info exists suspension] &&
		     [info exists channel])} {
			if {$s} {
				if {[catch {suspension_representation $channel} ev]} {
					set ev ""
				} else {
					set ev_is_suspension 1
					set ev_kind [set kind]suspension
					set ev_channel $channel
				}
			}
		}
	} else {
		if {[catch {event_kind $ev} kind_chan_given]} {
			return [list 0 "A_ERROR UnknownIOKind: event=$ev"]
		}
		set ev_kind  [lindex $kind_chan_given 0]
		set ev_channel [lindex $kind_chan_given 1]
		switch -- $ev_kind {
			input -
			output {
				set ev_is_suspension 0
			} inputsuspension -
			  outputsuspension {
				set ev_is_suspension 1
			}
		}
	}
	if {[info exists channel] &&
	    ![info exists ev_is_suspension]} {
		if {![catch {suspension_representation $channel} drepr ]} {
			if {[eq_event $ev $drepr]} {
				set ev_is_suspension 1
				set ev_kind [set kind]suspension
			} else {
				set ev_is_suspension 0
			}
		} else {
			set ev_is_suspension 0
		}
	}
	if {[info exists suspension] &&
	    [info exists ev_is_suspension]} {
		if {$ev_is_suspension != $suspension} {
			return [list 0 "A_ERROR Inconsistency: event and suspension fields contradict each other"]
		}
	}
	if {[info exists channel] &&
	    [info exists ev_channel]} {
		if {[string compare $channel $ev_channel] != 0} {
			return [list 0 "A_ERROR Inconsistency: event and channel fields contradict each other"]
		}
	}
	if {[info exists ev_channel] &&
	    [info exists ev_kind] &&
	    [info exists ev_is_suspension]} {
		if {$ev_is_suspension} {
			return [list 1 $ev_kind $ev $ev_channel $ev_is_suspension]
		} elseif {[string compare $ev ""] != 0} {
			return [list 1 $ev_kind $ev $ev_channel $ev_is_suspension]
		} else {
			return [list 0 "A_ERROR ArgumentMissing: Usage: $cmd event-to-be-checked"]
		}
	} else {
		return [list 0 "A_ERROR ArgumentMissing: Usage: $cmd event-to-be-checked"]
	}
}

proc primer_do_iokind {kind_sug chan_sug} {
	global susp_history
	# debugMsg "primer_do_iokind ($kind_sug) ($chan_sug)"
	if {[string compare $kind_sug ""] != 0} {
		if {![is_iokind $kind_sug]} {
			set result "A_ERROR WrongValue: not an iokind: \"$kind_sug\""
		}
	}
	if {[string compare $chan_sug ""] != 0} {
		if {![is_channel $chan_sug]} {
			set result "A_ERROR WrongValue: not a channel: \"$chan_sug\""
		}
	}
	if {([string compare $chan_sug ""] != 0) &&
            ([string compare $kind_sug ""] == 0)} {
		set kind_sug [get_kind $chan_sug]
	} elseif {![info exists result]} {
		if {([string compare $kind_sug ""] != 0) &&
		    ([string compare $chan_sug ""] != 0)} {
			if {[string compare $kind_sug [get_kind $chan_sug]] != 0} {
				set result "A_ERROR Inconsistency: channel and iokind fields conflict"
			}
		}
	}
	if {![info exists result]} {
		set havekind(input) [primer_menu_contains [list input inputsuspension]]
		set havekind(output) [primer_menu_contains [list output outputsuspension ]]
		# debugMsg  "primer_do_iokind in($havekind(input)) out($havekind(output))"
	}
	if {![info exists result]} {
		if {(!$havekind(input)) && (!$havekind(output))} {
			set result "A_IOKIND iokind=stop"
		}
	}
	if {![info exists result]} {
		if {([string compare $kind_sug ""] != 0) &&
		     $havekind($kind_sug)} {
			if {[string compare $chan_sug ""] != 0} {
		    		# FIXME should check if channel sug is available in menu
				set result "A_IOKIND iokind=$kind_sug\tchannel=$chan_sug"
			} else {
				set chan [primer_choose_channel $kind_sug]
				set result "A_IOKIND iokind=$kind_sug\tchannel=$chan"
			}
		} else {
			# debugMsg "primer_do_iokind ($susp_history(output)) == ([get_channels outputsuspension])"
			if {!$havekind(input)} {
			    	# FIXME schoose_channel should only choose what is available in menu
				set chan [primer_choose_channel output]
				set result "A_IOKIND iokind=output\tchannel=$chan"
			} elseif {!$havekind(output)} {
			    	# FIXME schoose_channel should only choose what is available in menu
				set chan [primer_choose_channel input]
				set result "A_IOKIND iokind=input\tchannel=$chan"

			} elseif {[llength $susp_history(output)] ==
			          [llength [get_channels outputsuspension]]} {
			    	# FIXME schoose_channel should only choose what is available in menu
				set chan [primer_choose_channel input]
				set result "A_IOKIND iokind=input\tchannel=$chan"
			} else {
				# FIXME: we still may have to avoid suspension on one or more channels!
				# we keep a list of channels on which we
				# have recently seen suspension!
				# randomly choose between input and output
				set kind [primer_choose_kind ]
				set chan [primer_choose_channel $kind]
				set result "A_IOKIND iokind=$kind\tchannel=$chan"
			}
		}
	}
	debugMsg "primer_do_iokind ($kind_sug) ($chan_sug): $result"
	puts $result
}

proc primer_do_channels {kind_sug} {
	if {[string compare $kind_sug ""] != 0} {
		if {![is_iokind $kind_sug]} {
			set result "A_ERROR WrongValue: not an iokind: \"$kind_sug\""
		} else {
			set kinds [list $kind_sug ]
		}
	} else {
		set kinds [list input output]
	}
	if {![info exists result]} {
		puts "A_CHANNELS_START"
		foreach k $kinds {
			set cl [get_channels $k]
			foreach c $cl {
				puts "A_CHANNEL iokind=$k\tchannel=$c"
			}
		}
		puts "A_CHANNELS_END"
	} else {
		puts $result
	}
}


proc primer_menu_contains {kinds} {
	foreach k $kinds {
		if {[llength [mr_list [list $k]]] > 0} {
			return 1
		}
	}
	return 0
}

# primer_choose_next:
# choose next action based on the given fields; return either
#    [list 1 kind ev chan s] (succesfull choice)
#    [0 errmsg]              (unsufficient or incomplete info)
#
# Assumptions:
#  - we either generate or recognize suspension-events ((m)ioco resp traces)
#  - suspension-events may have an empty-string representation;
#    this should not pose problems, because we can describe them
#    using fields chan and s.
#  - each given field may be emtpy string; this only poses problems
#    if the other fields do not contain sufficient information
#    to have a result.
# NOTE: the returned kind will be one of:
#  input, output, inputsuspension, outputsuspension
proc primer_choose_next { kind ev chan s cmd } {
	switch -- $kind {
		input -
		output {
		} default {
			return [list 0 "A_ERROR InternalError: primer_choose_next: kind notin (input,output): \"$kind\""
		}
	}
	if {[string compare $chan ""] != 0} {
		if {[is_channel $chan]} {
			set channel $chan
		} else {
			return [list 0 "A_ERROR WrongValue: not a channel: \"$chan\""]
		}
	}
	if {[string compare $s ""] != 0} {
		if {$s} {
			set suspension 1
		} else {
			set suspension 0
		}
	}
	set evs [concat [mr_list $kind $chan] [mr_list [set kind]suspension $chan]]
	debugMsg "primer_choose_next ($evs)"
	set l [llength $evs]
	if {$l == 0} {
		regsub "^C_" $cmd A_ rcmd
		if {[string compare $chan ""] != 0} {
			set forchan " for channel $chan"
		} else {
			set forchan ""
		}
		return [list 0 "[set rcmd]_ERROR reason=empty $kind menu$forchan"]
	}
 	set r [random $l]
	set el [lindex $evs $r]
	set el_longkind [lindex $el 0]
	set el_chan [lindex $el 1]
	set el_ev [lindex $el 2]
	switch -- $el_longkind {
		inputsuspension -
		outputsuspension {
				set el_s 1
		} input -
		  output {
				set el_s 0
		} default {
				set el_s ""
		}
	}
	return [list 1 $el_longkind $el_ev $el_chan $el_s]
}

proc primer_choose_channel { {kind ""} } {
	if {[string compare $kind ""] == 0} {
		set c [concat [get_channels input] [get_channels output]]
	} else {
		set c [get_channels $kind]
	}
	set l [llength $c]
	set r [random $l]
	return [lindex $c $r]
}

proc primer_choose_kind { {chan ""} } {
	if {[string compare $chan ""] == 0} {
		if {[random 2]} {
			set result input
		} else {
			set result output
		}
	} else {
		return [get_kind $chan]
	}
}

proc smile_event2event {ev} {
	set tmp ""
	append tmp [lindex $ev 0] ;# the gate name
	set exps [lindex $ev 1]
	set preds [lindex $ev 2]
	foreach exp $exps {
		set question_exclam [lindex $exp 0] ;# the ! or ?
		if {[string compare $question_exclam "?"] == 0} {
			append tmp " !"
		} else {
			append tmp " $question_exclam"
		}
		append tmp " [lindex $exp 1]" ;# the expression
	}
	return $tmp
}

proc send_menu_verdict {} {
	global menuinfo
	if {[string compare $menuinfo(verdict) ""] != 0} {
		puts "A_GUIDANCE_VERDICT $menuinfo(verdict)"
	} elseif {$menuinfo(inconsistent)} {
		puts "A_GUIDANCE_VERDICT INCONSISTENT"
	}
}

proc set_menu_verdict {s} {
	global menuinfo
	lappend menuinfo(verdict) $s
}

proc reset_menu_verdict {} {
	global menuinfo
	set menuinfo(verdict) {}
	set menuinfo(inconsistent) 0
}


proc primer_store_super {} {
	global cache
	
	set key [lindex [primer_get_super] 0]
	if {![info exists cache($key)]} {
		set cache($key) [mr_get]
	} else {
		debugMsg "\tprimer_store_super already $key"
	}
}

proc primer_retr_super {key} {
	global cache
	
	mr_set $cache($key)
}


proc primer_get_super {} {
	global superstate
	return $superstate(cur)
}

# this proc should be called _after_ obs_menu, because obs_menu
# updates the mapping table between 'new' and 'identical' states
# for the states in the 'current' superstate
# S1 = init, S2 = trans
proc primer_set_super {init trans ign id} {
	global superstate
	if {![info exists superstate(cur)]} {
		set superstate(cur) {}
	}
	debugMsg "primer_set_super ($superstate(cur)) $init $trans $ign"
	foreach i $ign {
		set excl($i) 1
	}
	set S1 {}
	foreach i $init {
		if {![info exists excl($i)]} {
			lappend S1 $i
		}
	}
	set S2 {}
	foreach i $trans {
		if {![info exists excl($i)]} {
			lappend S2 $i
		}
	}
	set canonical1 [expansion_canonicalize $S1]
	set seq1 [primer_mksupersequence $canonical1]
	set canonical2 [expansion_canonicalize $S2]
	set seq2 [primer_mksupersequence $canonical2]
	set seq3 [concat $canonical1 $canonical2]
	set idnr [primer_get_super_id $seq3 $init $trans $ign $S1 $S2 $seq1 $seq2 $id]
	set superstate(cur) [list $idnr $seq1 $seq2 $id]
	debugMsg "primer_set_super $init $trans -> $superstate(cur)"
}

# we number the canonical superstates
# (check if we have seen it before, if so, return its number,
#  or else, give it the next free superstate number, increment it,
#  and return the number we just dealt out)
proc primer_get_super_id {S init trans ign S1 S2 seq1 seq2 ident} {
	global superstate
	# first canonicalize:
	# (assuming expansion_canonicalize has been run already)
	# - strip edge/location info
	# - remove duplicates
	# - sort numerically
	set R {}
	foreach s $S {
		set r [strip_all_except_state $s]
		if {![info exists there($r)]} {
			lappend R $r
			set there($r) 1
		}
	}
	set RR [lsort -integer $R]
	if {![info exists superstate(id,$RR)]} {
		if {![info exists superstate(nextid)]} {
			set superstate(nextid) 0
		}
		set id $superstate(nextid)
		set superstate(id,$RR) $id
		set superstate(set,$id) $RR
		set superstate(init,$id) $init
		set superstate(trans,$id) $trans
		set superstate(ign,$id) $ign
		set superstate(seq1,$id) $seq1
		set superstate(seq2,$id) $seq2
		set superstate(ident,$id) $ident
		incr superstate(nextid)
	}
	if {$S1 != {}} {
		primer_set_init_id $superstate(id,$RR) $seq1
	}
	return $superstate(id,$RR)
}

proc primer_set_init_id {idnr S} {
	global superstate
	set R {}
	foreach s $S {
		set r [strip_all_except_state $s]
		if {![info exists there($r)]} {
			lappend R $r
			set there($r) 1
		}
	}
	set RR [lsort -integer $R]
	if {![info exists superstate(revinit,$RR)]} {
		set superstate(revinit,$RR) $idnr
	} elseif {[lsearch $superstate(revinit,$RR) $idnr] < 0} {
		lappend superstate(revinit,$RR) $idnr
	}
}

proc primer_get_init_id {S} {
	global superstate
	set R {}
	foreach s $S {
		set r [strip_all_except_state $s]
		if {![info exists there($r)]} {
			lappend R $r
			set there($r) 1
		}
	}
	set RR [lsort -integer $R]
	if {![info exists superstate(revinit,$RR)]} {
		return {}
	} else {
		return [lindex $superstate(revinit,$RR) 0]
	}
}

proc primer_goto_super_id {idnr} {
	global superstate

	if {[string compare $idnr ""] == 0} {
		puts "A_ERROR ArgumentMissing: Usage: C_GOTO super-id"
		return
	}

	if {[info exists superstate(set,$idnr)]} {
		set init $superstate(init,$idnr)
		set transitive $superstate(trans,$idnr)
		set all $superstate(set,$idnr)
		set ident $superstate(ident,$idnr)
		if {0} {
			set transitive [lindex [obs_menu $init 1 $ident] 1]
		} else {
			primer_retr_super $idnr
		}		
		set seq1 $superstate(seq1,$idnr)
		set seq2 $superstate(seq2,$idnr)
		set superstate(cur) [list $idnr $seq1 $seq2 $ident]
		puts "A_GOTO_OK"
	} else {
		puts "A_GOTO_ERROR UnknownStateID super=$idnr"
	}
}

proc primer_strip_edge_super {} {
	global superstate
debugMsg "super-old: $superstate(cur)"
	set idnr [lindex $superstate(cur) 0]
	set seq1orig [lindex $superstate(cur) 1]
	set seq2orig [lindex $superstate(cur) 2]
	set ident [lindex $superstate(cur) 3]
	set seq1upd {}
	foreach e [split $seq1orig ,] {
		lappend seq1upd [strip_edge $e]
	}
	set seq1 [join $seq1upd ,]
	# set seq2upd {}
	# foreach e [split $seq2orig ,] {
	# 	lappend seq2upd [strip_edge $e]
	# }
	# set seq2 [join $seq2upd ,]
	set seq2 $seq2orig
	set superstate(cur) [list $idnr $seq1 $seq2 $ident]
debugMsg "super-new: $superstate(cur)"
}

proc primer_mksupersequence {S} {
	# For now, just return the sorted result
	set sorted [lsort $S]
	return [join $sorted ,]
	# for sequences of numbers, replace things like 4,5,6,7 by 4-7 etc.
	set sorted [lsort -integer $S]
	set result ""
	if {[llength $sorted] > 0} {
		set last [lindex $sorted 0]
		append result $last
		set in_sequence 0
		foreach e [lrange $sorted 1 end] {
			if {$e == [expr $last + 1]} {
				set in_sequence 1
			} else {
				if {$in_sequence} {
					set in_sequence 0
					append result "-$last"
				}
				append result ",$e"
			}
			set last $e
		}
		if {$in_sequence} {
			append result "-$last"
		}
	}
	return $result
}
