/***** trojka: tan-sman.c *****/

/* Copyright (c) 1991,1995 by AT&T Corporation.  All Rights Reserved.     */
/* This software is for educational purposes only.                        */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.att.com          */

/*	Modified code and additional routines for the trojka analyser	*/
/*	No guarantee is expressed or implied!				*/
/*									*/
/* No permission is given to distribute the modified code and		*/
/* additional routines without permission of the author.		*/
/*									*/
/* Modified code and additional routines written by Rene G. de Vries	*/
/*									*/
/* trojka analyser state manager					*/
/* version 1.0 :	Tue Jun 11 20:32:02 MET DST 1996		*/
/*----------------------------------------------------------------------*/

/* === declaration of hidden datatypes and variables === */

#include <stdio.h>

#ifdef TANSTATIC
#include "tan-sman.h"	
#endif

StateSetT *ssbag        = NULL; /* global bag of unused statevectors */
StateSetT **T_tab       = NULL; /* Global hashtable of statevectors */
StateStrT *stbag	= NULL; /* Global bag of unused state-structures */ 
StateStrT *stroot	= NULL; /* root of statestructure link */

int createdstates	= 0;	/* number of created slots to store */
				/* statevectors */
int createdstructs	= 0;	/* number of created slots to store */
				/* structures of statevectors */
/* === procedures and functions (private aswell public) === */

StateSetT *
SSfree() /* return a free statevector from ssbag */
{	StateSetT *sstate;
	if (!ssbag)	/* none, so create one */	
	{	sstate = (StateSetT *) emalloc(sizeof(StateSetT));
 		sstate->body = (char *) emalloc(sizeof(State));  /* maximum */ 

		createdstates++; /* count number of created vectors */
	}
	else {	/* well, there's something left */
		sstate = ssbag;
		ssbag = ssbag->nxt;
		sstate->nxt = NULL;
		sstate->vsize = 0;
	}
	return sstate;
}

StateStrT *
STfree() /* return a free structure from stbag */
{	StateStrT *sstruc;
	if (!stbag)	/* none, so create one */	
	{	sstruc = (StateStrT *) emalloc(sizeof(StateStrT));
		createdstructs++;
	}
	else {	/* well, there's something left */
		sstruc = stbag;
		stbag = stbag->nxt;
	}
	sstruc->nxt  = NULL;	/* defaults values */
	sstruc->last = NULL;
	sstruc->nr_users = 0;
	return sstruc;
}
			
int
SSismember(StateSetT **superstate, State *state, int size, char *p, char *q,
	   uchar *delta)
/* return if state is member of the supervector state and corsponding delta*/
{	StateSetT *sstate;
	int m;

	for (sstate=*superstate;sstate;sstate = sstate->nxt) {
	m = memcmp((char *)sstate->body,(char *)state,size);
		if (m==0 && size==sstate->vsize &&
		   (memcmp((char *) sstate->struc->proc_struct, p, \
			(state->_nr_pr)*sizeof(int)) == 0) &&
		   (memcmp((char *) sstate->struc->q_struct, q, \
			(state->_nr_qs)*sizeof(int))==0))
		   {	*delta = sstate->delta;
			return 1; /* is member */
		   }
		else if (m<0) break;
	 }
	 return 0;
}

void
SSaddstate(	StateSetT **superstate,	/* io */ 
	 	State *state,		/* in */
	 	int size,		/* in */
		char *p,		/* in: proc_offset */
		char *q,		/* in: q_offset */
		uchar delta		/* in: delta */
	)
{	StateSetT *tmp, *ostate = NULL, *ntmp;
	StateStrT *sstruc, *ostruc = NULL, *nstruc;
	int m;

	if (!*superstate) *superstate = tmp = SSfree();
	else {
	    for (tmp=*superstate;tmp;ostate = tmp, tmp = tmp->nxt) {
	    m = memcmp((char *)tmp->body,(char *)state,size);
		if (m==0 && size==tmp->vsize && \
		   (memcmp((char *) tmp->struc->proc_struct, p, \
			(state->_nr_pr)*sizeof(int)) == 0) && \
		   (memcmp((char *) tmp->struc->q_struct, q, \
			(state->_nr_qs)*sizeof(int))==0))
			return; /* already a member */
		else if (m<0) 
		{	ntmp = SSfree();
			ntmp->nxt = tmp;
			if (!ostate) *superstate = ntmp;
			else	ostate->nxt = ntmp;
			tmp = ntmp;
			break;		
		}
		else if (!tmp->nxt)
		{	tmp->nxt = SSfree();
			tmp = tmp->nxt;
			break;
		}
	     }
	}
	tmp->vsize = size;	/* copy the items */
	tmp->delta = delta;
	memcpy((char *)(tmp->body), state, size);
	if (!stroot) stroot = sstruc = STfree();
	else
	for (sstruc=stroot;sstruc;ostruc = sstruc, sstruc = sstruc->nxt) {

	m = memcmp((char *) (sstruc->proc_struct), p,
		(state->_nr_pr)*sizeof(int));
	if (m==0 && \
	    memcmp((char *) (sstruc->q_struct), q, 
		(state->_nr_qs)*sizeof(int)) == 0) 
		{	tmp->struc = sstruc;	 /* is member */
			sstruc->nr_users += 1;
			return;
		}
	else if (m<0) 
	{	nstruc = STfree();
		nstruc->nxt  = sstruc;
		nstruc->last = sstruc->last;	/* beware: double linked! */
		sstruc->last = nstruc;
		if (!ostruc) stroot = nstruc;
		else ostruc->nxt = nstruc;
		sstruc = nstruc;
		break;
	}
	else if (!sstruc->nxt)
	{	sstruc->nxt  = STfree();
		sstruc->nxt->nxt = NULL;
		sstruc->nxt->last = sstruc;
		sstruc = sstruc->nxt;
		break;
	}
	} /* all possible members */
	tmp->struc = sstruc;
	memcpy((char *) tmp->struc->proc_struct, p, MAXPROC*sizeof(int));
	memcpy((char *) tmp->struc->q_struct, q, MAXQ*sizeof(int));
	tmp->struc->nr_users += 1;
} /*addstate*/

void
SSclear (StateSetT **superstate)
/* empty superstate and release all statevectors to everybody */ 
{	StateSetT *sstate, *ostate = NULL;			
	if (*superstate) {	/* anything to do? */
		for (sstate=*superstate;(sstate);ostate=sstate,
					sstate=sstate->nxt) 
		if(! --(sstate->struc->nr_users)) {
			/* we may dispose the structure info */
			if (!sstate->struc->last)	/* stroot? */
			   stroot = sstate->struc->nxt;
			else 			
			   sstate->struc->last->nxt = sstate->struc->nxt;
			if (sstate->struc->nxt) /* not last item of chain */
			   sstate->struc->nxt->last = sstate->struc->last;
			sstate->struc->nxt = stbag;	/* release to stbag */
			stbag = sstate->struc;
			}
	 	ostate->nxt=ssbag;  /* release to free space */
	 	ssbag = *superstate;
	 	*superstate = NULL;
		}
}


void
SSmerge(StateSetT **Dsuperstate,	/* io */
	StateSetT **superstate		/* io */)
/* merge superstate to Dsuperstate, superstate will be empty */
{	StateSetT *sstate, *tmp;
	uchar dummy;

	if (*superstate) {	/* something to do */
	 sstate = *superstate;
	 while (sstate) {
		tmp = sstate;
		sstate=sstate->nxt;
		/* doubles will be disposed by SSaddstate */
	 	SSaddstate(Dsuperstate, (State *) tmp->body,
				tmp->vsize,
				(char *) tmp->struc->proc_struct,
				(char *) tmp->struc->q_struct,
				tmp->delta);
		tmp->nxt = NULL;
		SSclear(&tmp);
		}
	*superstate = NULL;
	}
}

void ResState(StateSetT*);

int
SSnrel (StateSetT **superstate)
/* return the number of elements in superstate*/ 
{	StateSetT *sstate;
	int count =0,i;			
	if (*superstate) 	/* anything to do? */
	 for (sstate=*superstate;sstate;sstate=sstate->nxt)
		{ count++;
		}
	return count;
}	

void
Tinit(void)
{
T_tab = (StateSetT **) emalloc((1<<ssize) * sizeof(StateSetT *));
}

void 
Trelease()
/*release all used space to ssbag */
{ int i;
  for(i=0;i<=mask;i++) SSclear(&T_tab[i]);
  truncs = 0;
}

void
Tstore(State *vin /* state */, int nin /*size*/, int *p, int *q, uchar delta) 
{
	s_hash((uchar *)vin, nin);
	SSaddstate(&T_tab[j1], vin, nin, (char *) p, (char *) q, delta);

}


int
Thit(State *vin /* state */, int nin /*size*/, int *p, int *q, uchar *delta) 
{
	s_hash((uchar *)vin, nin);
	return SSismember(&T_tab[j1], vin, nin, (char *) p, (char *) q, delta);
}


void    
IncludeState(StateSetT **Set)   /* saves state and context in Set  */
{
        SSaddstate(Set, &now, vsize, (char *) proc_offset,
                         (char *) q_offset, 0 /*DUMMY*/);
}

void
ResState(StateSetT *State)      /* restores state State in now ect. */
/* restore State */
{       if (State) {
                vsize = State->vsize;
                memcpy((char *) &now, State->body, vsize);
                memcpy((char *) proc_offset, State->struc->proc_struct,
                        MAXPROC*sizeof(int));
                memcpy((char *) q_offset, State->struc->q_struct,
                        MAXQ*sizeof(int));
        }
}
