/***** trojka: trojka.h *****/

/* Copyright (c) 1991,1995 by AT&T Corporation.  All Rights Reserved.     */
/* This software is for educational purposes only.                        */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.att.com          */

/*	Modified code and additional routines for the trojka analyser	*/
/*	No guarantee is expressed or implied!				*/
/*									*/
/* No permission is given to distribute the modified code and		*/
/* additional routines without permission of the author.		*/
/*									*/
/* Modified code and additional routines written by Rene G. de Vries	*/
/*									*/
/* trojka interface							*/
/* version 1.0 :	Tue Jun 11 20:32:02 MET DST 1996		*/

#ifndef H_TROJKA
#define H_TROJKA

#include <stdio.h>

/* exported datatypes */

typedef	enum	{ input, output } AtypeT;
typedef enum	{ variable, solid } fieldT;

typedef struct ActionT{	/* datatype to store an action */ 

	AtypeT	atype;		/* actiontype	*/
	short	Queue;		/* queue number */
	short	nr_fields;	/* nr of field */
	fieldT	*sfield;	/* field settting tag (array[1..nr_fields] ) */
	int	*vfield;	/* field value (array[1..nr_fields]) */
	
} ActionT;

typedef struct Trojka_SettingsT {	/* datatype to pass Trojka settings */

	FILE	*logfile;	/* file to store log-info */
	FILE	*searchfile;	/* file to store search-traces */
	int	hashnumber;	/* hash number to be used (see spin) */
	int	maxdepth;	/* maxium depth until cut is preformed */
	int	ssize;		/* hashtable statespace size: 2^ssize */  
	int	seed;		/* seed value for the random generator */

} Trojka_SettingsT;

typedef struct	Trojka_statT {		/* datatype to store statistics */

	/* Global parameters */
	int	nr_vector;	/* number of created statevectors */
	int	nr_svector;	/* number of created structures for vectors */
	double	memcnt;		/* memory usage counter */

	/* Parameters after each trojka_engine command */
	int	nr_cuts;	/* number of cut branches */
	int	nr_analysed;	/* number of actual analysed states */
	int	nr_matched;	/* number of matched states */
	int	nr_supers;	/* number of elements of superstate vector */
	int	nr_executed;	/* number of executed actions */
	int	depth;		/* maximum reached depth of branchs */
	char	delta;		/* current superstate has delta suspension */
	short	nr_actions;	/* number of possible actions */

} Trojka_statT;

/* exported procedures and functions */

void
Init_Trojka( Trojka_SettingsT /* in */ );	/* installs trojka_engine */

Trojka_statT
getstats(void);					/* return trojka statistics */

void
Analyse(void);		/* analyses the current superstate */

                               
void
PrtAction(FILE* , int /* actionnr */);
/* print action actionnr in text to output file out */


ActionT
GetAction(int);
/* return action identified by the parameter */

ActionT
SetAction(int);
/* instatntiate action identified by the parameter and return it */


void
ExecDelta();		/* pass an execute delta-move to the trojka-engine */

 
ActionT
ExecMove(int /* move */);	/* execute action move */

char*
Mval2Name(unsigned char /* mtype value */); /* give the symbolic name of a enumerated mtype */

unsigned char
MName2Val(char* /* mtype symbolic name */); /* give the enumerate number identifier of the symbolic name of a mtype member*/

char*
Qid2Name(int /* queue number */); /* give the symbolic name of queue cf. ActionT->Queue */

char**
QidFieldTypes(int /* queue number */); /* give types of the queue elements */

void
logprintf(char *, ...); /* print to the logfile logfile, used in setting printflike */

#endif
